from typing import TYPE_CHECKING, List, Optional, Tuple, cast

from loguru import logger

if TYPE_CHECKING:
    from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine
    from pymatgen.electronic_structure.dos import CompleteDos
    from pymatgen.electronic_structure.plotter import BSDOSPlotter


def average_along_axis(
    datafile: str = "potential.h5",
    task: str = "potential",
    axis: int = 2,
    smooth: bool = False,
    smooth_frac: float = 0.8,
    raw: bool = False,
    subtype: Optional[str] = None,
    verbose: bool = False,
    **kwargs,
):
    r"""Plot the average curve of a physical quantity along a certain axis

    Parameters
    ----------
    datafile:
        Path to an h5 or json file, or a folder containing any of these files, default 'potential.h5'
    task:
        Task type, can be 'rho', 'potential', 'elf', 'pcharge', 'rhoBound'
    axis:
        Along which axis to plot the potential curve, default is 2
    smooth:
        Whether to smooth, default False
    smooth_frac:
        Smoothing coefficient, default 0.8
    raw:
        Whether to return plot data to a CSV file
    subtype:
        Used to specify the task data subtype, default None, representing drawing Potential/TotalElectrostaticPotential
    **kwargs:
        Other parameters, passed to matplotlib.pyplot.plot

    Returns
    -------
    axes:
        Can be passed to other functions for further processing

    Examples
    --------
    >>> from dspawpy.plot import average_along_axis

    Read data from the potential.h5 file, plot, and save the original plot data to a CSV file

    >>> plt = average_along_axis(datafile='tests/3.3/rho.h5', task='rho', axis=2, smooth=True, smooth_frac=0.8) # doctest: +ELLIPSIS
    >>> plt.savefig('tests/outputs/doctest/rho_h5.png')
    >>> plt = average_along_axis(datafile='tests/3.3/rho.json', task='rho', axis=2, smooth=True, smooth_frac=0.8) # doctest: +ELLIPSIS
    >>> plt.savefig('tests/outputs/doctest/rho_json.png')

    >>> plt = average_along_axis(datafile='tests/2.7/potential.h5', task='potential', axis=2, smooth=True, smooth_frac=0.8, raw=True) # doctest: +ELLIPSIS
    >>> plt.savefig('tests/outputs/doctest/potential_h5.png')
    >>> plt = average_along_axis(datafile='tests/2.7/potential.json', task='potential', axis=2, smooth=True, smooth_frac=0.8) # doctest: +ELLIPSIS
    >>> plt.savefig('tests/outputs/doctest/potential_json.png')

    >>> plt = average_along_axis(datafile='tests/2.8/elf.h5', task='elf', axis=2, smooth=True, smooth_frac=0.8) # doctest: +ELLIPSIS
    >>> plt.savefig('tests/outputs/doctest/elf_h5.png')
    >>> plt = average_along_axis(datafile='tests/2.8/elf.json', task='elf', axis=2, smooth=True, smooth_frac=0.8) # doctest: +ELLIPSIS
    >>> plt.savefig('tests/outputs/doctest/elf_json.png')

    >>> plt = average_along_axis(datafile='tests/2.9/pcharge.h5', task='pcharge', axis=2, smooth=True, smooth_frac=0.8) # doctest: +ELLIPSIS
    >>> plt.savefig('tests/outputs/doctest/pcharge_h5.png')
    >>> plt = average_along_axis(datafile='tests/2.9/pcharge.json', task='pcharge', axis=2, smooth=True, smooth_frac=0.8) # doctest: +ELLIPSIS
    >>> plt.savefig('tests/outputs/doctest/pcharge_json.png')

    >>> plt = average_along_axis(datafile='tests/2.28/rhoBound.h5', task='rhoBound', axis=2, smooth=True, smooth_frac=0.8) # doctest: +ELLIPSIS
    >>> plt.savefig('tests/outputs/doctest/rhoBound_h5.png')
    >>> plt = average_along_axis(datafile='tests/2.28/rhoBound.json', task='rhoBound', axis=2, smooth=True, smooth_frac=0.8) # doctest: +ELLIPSIS
    >>> plt.savefig('tests/outputs/doctest/rhoBound_json.png')

    """
    assert task in [
        "rho",
        "potential",
        "elf",
        "pcharge",
        "rhoBound",
    ], "Only support: rho, potential, elf, pcharge, rhoBound"

    # only for compatibility
    import numpy as np

    if isinstance(datafile, (list, np.ndarray)):
        ys = datafile  # expect np.ndarray or list
    else:
        from dspawpy.io.utils import get_absfile

        absfile = get_absfile(datafile, task, verbose=verbose)

        if absfile.endswith(".h5"):
            hfile = absfile
            from dspawpy.io.read import load_h5

            hdict = load_h5(hfile)
            grid = hdict["/AtomInfo/Grid"]

            if task == "rho":
                if subtype is None:
                    subtype = "TotalCharge"
                _key = f"/Rho/{subtype}"
            elif task == "potential":
                if subtype is None:
                    if "/Potential/TotalElectrostaticPotential" in hdict:
                        subtype = "TotalElectrostaticPotential"
                    elif "/Potential/TotalLocalPotential" in hdict:
                        subtype = "TotalLocalPotential"
                    else:
                        raise KeyError(
                            f"Neither TotalElectrostaticPotential nor TotalLocalPotential can be found, please check your {absfile}",
                        )
                _key = f"/Potential/{subtype}"
            elif task == "elf":
                if subtype is None:
                    subtype = "TotalELF"
                _key = f"/ELF/{subtype}"
            elif task == "pcharge":
                if subtype is None:
                    subtype = "TotalCharge"
                _key = f"/Pcharge/1/{subtype}"
            elif task == "rhoBound":
                if subtype is None:
                    subtype = "TotalCharge"
                _key = f"/Rho/{subtype}"
            else:
                raise ValueError("Only support rho, potential, elf, pcharge, rhoBound")

            if _key not in hdict:
                raise KeyError(f"No {_key} key")

            tmp_pot = np.asarray(hdict[_key]).reshape([-1, 1], order="C")
            ys = tmp_pot.reshape(grid, order="F")

        elif absfile.endswith(".json"):
            jfile = absfile
            with open(jfile) as f:
                from json import load

                jdict = load(f)
            grid = jdict["AtomInfo"]["Grid"]

            if task == "rho":
                if subtype is None:
                    subtype = "TotalCharge"
                ys = np.asarray(jdict["Rho"][subtype]).reshape(grid, order="F")
            elif task == "potential":
                if subtype is None:
                    if "TotalElectrostaticPotential" in jdict["Potential"]:
                        subtype = "TotalElectrostaticPotential"
                    elif "TotalLocalPotential" in jdict["Potential"]:
                        subtype = "TotalLocalPotential"
                    else:
                        raise KeyError(
                            f"Neither TotalElectrostaticPotential nor TotalLocalPotential can be found, please check your {absfile}",
                        )

                ys = np.asarray(jdict["Potential"][subtype]).reshape(grid, order="F")
            elif task == "elf":
                if subtype is None:
                    subtype = "TotalELF"
                ys = np.asarray(jdict["ELF"][subtype]).reshape(grid, order="F")
            elif task == "pcharge":
                if subtype is None:
                    subtype = "TotalCharge"
                ys = np.asarray(jdict["Pcharge"][0][subtype]).reshape(grid, order="F")
            else:
                if subtype is None:
                    subtype = "TotalCharge"
                ys = np.asarray(jdict["Rho"][subtype]).reshape(grid, order="F")

        else:
            raise TypeError("Only support h5/json file")

    all_axis = [0, 1, 2]
    all_axis.remove(axis)
    y = np.mean(ys, tuple(all_axis))  # type: ignore
    x = np.arange(len(y))

    import matplotlib.pyplot as plt

    if raw:
        import polars as pl

        pl.DataFrame({"x": x, "y": y}).write_csv(
            f"raw{task}_axis{axis}.csv",
        )
    if smooth:
        import statsmodels.api as sm

        s = sm.nonparametric.lowess(y, x, frac=smooth_frac)
        if raw:
            import polars as pl

            pl.DataFrame({"x": s[:, 0], "y": s[:, 1]}).write_csv(
                f"raw{task}_axis{axis}_smooth.csv",
            )

        plt.plot(s[:, 0], s[:, 1], label="macroscopic average", **kwargs)

    plt.plot(x, y, **kwargs)

    return plt


def plot_aimd(
    datafile: str = "aimd.h5",
    show: bool = True,
    figname: str = "aimd.png",
    flags_str: str = "12345",
    raw: bool = False,
):
    r"""Plot the convergence process of key physical quantities after the AIMD task completion

    aimd.h5 -> aimd.png

    Parameters
    ----------
    datafile:
        Location of the h5 file. For example, 'aimd.h5' or ['aimd.h5', 'aimd2.h5']
    show:
        Whether to display the interactive interface. Default is False
    figname:
        Path to the saved image. Default 'aimd.h5'
    flags_str:
        Subplot number.
        1. Kinetic Energy
        2. Total Energy
        3. Pressure
        4. Temperature
        5. Volume
    raw:
        Whether to output plot data to a CSV file

    Returns
    -------
    figname:
        Image path, default 'aimd.png'

    Examples
    --------
    >>> from dspawpy.plot import plot_aimd

    Read the contents of the aimd.h5 file, plot the convergence process graphs of kinetic energy, total energy, temperature, and volume, and save the corresponding data to rawaimd_*.csv.

    >>> plot_aimd(datafile='tests/2.18/aimd.h5', flags_str='1 2 3 4 5', raw=True, show=False, figname="tests/outputs/doctest/aimdconv.png")
    >>> plot_aimd(datafile='tests/2.18/aimd.json', flags_str='1 2 3 4 5', show=False, figname="tests/outputs/doctest/aimdconv_json.png")

    """
    # Handle user input, deduplicate in order
    temp = set()
    flags = [x for x in flags_str if x not in temp and (temp.add(x) or True)]
    if " " in flags:  # remove space
        flags.remove(" ")

    for flag in flags:
        assert flag in ["1", "2", "3", "4", "5"], "flag must be in '12345'"

    if datafile.endswith("json"):
        # delete 3 from flags
        flags = [f for f in flags if f != "3"]
        logger.warning(
            "PressureKinetic array is not written in json file, so dspawapy can not read it!\nYou may try aimd.h5 file, which contains that info",
        )

    # Start plotting the combined figure
    N_figs = len(flags)
    import matplotlib.pyplot as plt

    fig, axes = plt.subplots(N_figs, 1, sharex=True, figsize=(6, 2 * N_figs))
    if N_figs == 1:  # 'AxesSubplot' object is not subscriptable
        axes = [axes]  # Avoid the type error mentioned above
    fig.suptitle("DSPAW AIMD")

    for i, flag in enumerate(flags):
        # Read data
        if i == 0:
            xs, ys = _read_aimd_converge_data(datafile, flag, verbose=True)
        else:
            xs, ys = _read_aimd_converge_data(datafile, flag, verbose=False)
        logger.info("For subfigure " + flag)

        if raw:
            import polars as pl

            pl.DataFrame({"x": xs, "y": ys}).write_csv(
                f"rawaimd_{flag}.csv",
            )

        axes[i].plot(xs, ys)  # Plot the coordinates
        # y-axis label for the subplot
        if flag == "1":
            axes[i].set_ylabel("Kinetic Energy (eV)")
        elif flag == "2":
            axes[i].set_ylabel("Energy (eV)")
        elif flag == "3":
            axes[i].set_ylabel("Pressure Kinetic (kbar)")
        elif flag == "4":
            axes[i].set_ylabel("Temperature (K)")
        else:
            axes[i].set_ylabel("Volume (Angstrom^3)")

    plt.tight_layout()
    # save and show
    if figname:
        import os

        absfig = os.path.abspath(figname)
        os.makedirs(os.path.dirname(absfig), exist_ok=True)
        plt.savefig(absfig, dpi=300)
        logger.info(f"==> {absfig}")
    if show:
        plt.show()


def plot_bandunfolding(
    datafile: str = "band.h5",
    ef: Optional[float] = None,
    de: float = 0.05,
    dele: float = 0.06,
    raw: bool = False,
    verbose: bool = False,
):
    r"""After the band unfolding task is completed, read data from h5 or json files to plot

    band.h5/band.json -> bandunfolding.png

    Parameters
    ----------
    datafile:
        Path to an h5 or json file or a directory containing any of these files, default 'band.h5'
    ef:
        Fermi level, default read from file /UnfoldingBandInfo/Efermi record data
    de:
        Band width, default 0.05
    dele:
        Band gap width, default 0.06
    raw:
        Whether to output plotting data to rawbandunfolding.csv

    Returns
    -------
    axes: matplotlib.axes._subplots.AxesSubplot
        Can be passed to other functions for further processing

    Examples
    --------
    Plot and save the plot data to rawbandunfolding.csv

    >>> from dspawpy.plot import plot_bandunfolding
    >>> plt = plot_bandunfolding("tests/2.22/band.h5", raw=True)
    >>> plt.savefig('tests/outputs/doctest/bandunfolding_h5.png')
    >>> plt = plot_bandunfolding("tests/2.22/band.json")
    >>> plt.savefig('tests/outputs/doctest/bandunfolding_json.png')

    """
    from dspawpy.io.utils import get_absfile

    absfile = get_absfile(datafile, task="band", verbose=verbose)
    import numpy as np

    if absfile.endswith(".h5"):
        from h5py import File

        f = File(absfile, "r")
        if ef is None:
            ef = np.asarray(f["/UnfoldingBandInfo/EFermi"])[0]
        number_of_band = np.asarray(f["/BandInfo/NumberOfBand"])[0]
        number_of_kpoints = np.asarray(f["/BandInfo/NumberOfKpoints"])[0]
        data = np.asarray(f["/UnfoldingBandInfo/Spin1/UnfoldingBand"])
        weight = np.asarray(f["/UnfoldingBandInfo/Spin1/Weight"])
    elif absfile.endswith(".json"):
        with open(absfile) as f:
            from json import load

            band = load(f)
        if ef is None:
            ef = band["UnfoldingBandInfo"]["EFermi"]
        number_of_band = band["BandInfo"]["NumberOfBand"]
        number_of_kpoints = band["BandInfo"]["NumberOfKpoints"]
        data = band["UnfoldingBandInfo"]["Spin1"]["UnfoldingBand"]
        weight = band["UnfoldingBandInfo"]["Spin1"]["Weight"]
    else:
        raise TypeError("Only support h5/json file")

    celtot = np.asarray(data).reshape((number_of_kpoints, number_of_band)).T
    proj_wt = np.asarray(weight).reshape((number_of_kpoints, number_of_band)).T
    X2, Y2, Z2, emin = getEwtData(
        number_of_kpoints,
        number_of_band,
        celtot,
        proj_wt,
        ef,
        de,
        dele,
    )

    if raw:
        import polars as pl

        pl.DataFrame({"Y": Y2, "Z": Z2}).write_csv("rawbandunfolding.csv")

    import matplotlib.pyplot as plt

    plt.clf()
    plt.scatter(X2, Y2, c=Z2, cmap="hot")
    plt.xlim(0, number_of_kpoints)
    plt.ylim(emin - 0.5, 15)
    ax = plt.gca()
    plt.colorbar()
    ax.set_facecolor("black")

    return plt


def plot_optical(
    datafile: str = "optical.h5",
    keys: List[str] = [
        "AbsorptionCoefficient",
        "ExtinctionCoefficient",
        "RefractiveIndex",
        "Reflectance",
    ],
    axes: List[str] = ["X", "Y", "Z", "XY", "YZ", "ZX"],
    raw: bool = False,
    prefix: str = "",
    save: bool = True,
    verbose: bool = False,
):
    """After the optical property calculation task is completed, read the data and draw a preview image

    optical.h5/optical.json -> optical.png

    Parameters
    ----------
    datafile:
        Path to an h5 or json file, or a folder containing any of these files, default 'optical.h5'
    keys:
        One of "AbsorptionCoefficient", "ExtinctionCoefficient", "RefractiveIndex", "Reflectance", default "AbsorptionCoefficient"
    axes:
        Index, default "X", "Y", "Z", "XY", "YZ", "ZX"
    raw:
        Whether to save plot data to CSV
    prefix:
        Folder path to save images, if empty, saves in the current directory
    save:
        Whether to save the image, default is True

    Examples
    --------
    Plot and save the plot data to rawoptical.csv

    >>> from dspawpy.plot import plot_optical
    >>> plot_optical("tests/2.12/scf.h5", "AbsorptionCoefficient", ['X', 'Y'], prefix='tests/outputs/doctest')
    >>> plot_optical("tests/2.12/optical.json", ["AbsorptionCoefficient"], ['X', 'Y'], prefix='tests/outputs/doctest', raw=True)

    """
    import matplotlib.pyplot as plt
    import numpy as np
    from scipy.interpolate import interp1d

    from dspawpy.io.utils import get_absfile

    if isinstance(keys, str):
        keys = [keys]

    absfile = get_absfile(datafile, task="optical", verbose=verbose)
    data = {}
    if absfile.endswith("h5"):
        from dspawpy.io.read import load_h5

        data_all = load_h5(absfile)
        energy = data_all["/OpticalInfo/EnergyAxe"]
        for k in keys:
            data[k] = data_all["/OpticalInfo/" + k]
    elif absfile.endswith("json"):
        with open(absfile) as fin:
            from json import load

            data_all = load(fin)
        energy = data_all["OpticalInfo"]["EnergyAxe"]
        for k in keys:
            data[k] = data_all["OpticalInfo"][k]
    else:
        raise TypeError("Only support h5/json file")

    energy_spline = np.linspace(energy[0], energy[-1], 2001)
    dd = {"X": 0, "Y": 1, "Z": 2, "XY": 3, "YZ": 4, "ZX": 5}
    data_spline_dict = {}

    for k, v in data.items():
        if save:
            plt.clf()
            # plt.ylabel(k + r" $\alpha (\omega )(cm^{-1})$", fontsize=14)
            plt.ylabel(k, fontsize=14)
        else:
            plt.ylabel("Properties", fontsize=14)
        plt.xlabel("Photon energy (eV)", fontsize=14)
        # TODO: unit
        for a in axes:
            d = np.asarray(v).reshape(len(energy), 6)[:, dd[a]]
            inter_f = interp1d(energy, d, kind="cubic")
            data_spline = inter_f(energy_spline)
            if save:
                plt.plot(energy_spline, data_spline, label=a)
            else:
                plt.plot(energy_spline, data_spline, label=" ".join([k, a]))
            data_spline_dict.update({k + a: data_spline})
        plt.legend()
        if save:
            if prefix:
                import os

                os.makedirs(prefix, exist_ok=True)
                plt.savefig(f"{prefix}/{k}.png")
            else:
                plt.savefig(f"{k}.png")

    if raw:
        import polars as pl

        for k in data:
            dictOriginal = {"energy": energy}
            dictInterp = {"energy": energy_spline}
            for a in axes:
                dictOriginal.update(
                    {k + a: np.asarray(data[k]).reshape(len(energy), 6)[:, dd[a]]},
                )
                dictInterp.update({k + a: data_spline_dict[k + a]})
            pl.DataFrame(dictOriginal).write_csv(
                f"rawoptical{k}.csv",
            )
            pl.DataFrame(dictInterp).write_csv(
                f"rawoptical_spline{k}.csv",
            )


def plot_phonon_thermal(
    datafile: str = "phonon.h5",
    figname: str = "phonon.png",
    show: bool = True,
    raw: bool = False,
    verbose: bool = False,
):
    """Task completed for phonon thermodynamic calculations, plot curves of relevant physical quantities versus temperature

    phonon.h5/phonon.json -> phonon.png

    Parameters
    ----------
    datafile:
        Path to an h5 or json file or a folder containing any of these files, default 'phonon.h5'
    figname:
        Filename to save the image
    show:
        Whether to pop up an interactive interface
    raw:
        Whether to save the plotting data to rawphonon.csv file

    Returns
    -------
    figname:
        Image path, default 'phonon.png'

    Examples
    --------
    >>> from dspawpy.plot import plot_phonon_thermal
    >>> plot_phonon_thermal('tests/2.26/phonon.h5', figname='tests/outputs/doctest/phonon_thermal_h5.png', show=False)
    >>> plot_phonon_thermal('tests/2.26/phonon.json', figname='tests/outputs/doctest/phonon_thermal_json.png', show=False, raw=True)

    """
    from dspawpy.io.utils import get_absfile

    absfile = get_absfile(datafile, task="phonon", verbose=verbose)

    import numpy as np

    if absfile.endswith(".h5"):
        hfile = absfile
        from h5py import File

        ph = File(hfile, "r")
        if "/ThermalInfo/Temperatures" not in ph:
            raise KeyError(
                f"No thermal info in {absfile}, you probably gave a wrong phonon.h5 file",
            )
        temp = np.asarray(ph["/ThermalInfo/Temperatures"])
        entropy = np.asarray(ph["/ThermalInfo/Entropy"])
        heat_capacity = np.asarray(ph["/ThermalInfo/HeatCapacity"])
        helmholts_free_energy = np.asarray(ph["/ThermalInfo/HelmholtzFreeEnergy"])
    elif absfile.endswith(".json"):
        jfile = absfile
        with open(jfile) as f:
            from json import load

            data = load(f)
        temp = np.asarray(data["ThermalInfo"]["Temperatures"])
        entropy = np.asarray(data["ThermalInfo"]["Entropy"])
        heat_capacity = np.asarray(data["ThermalInfo"]["HeatCapacity"])
        helmholts_free_energy = np.asarray(data["ThermalInfo"]["HelmholtzFreeEnergy"])
    else:
        raise TypeError("Only support h5/json file")

    if raw:
        import polars as pl

        pl.DataFrame(
            {
                "temp": temp,
                "entropy": entropy,
                "heat_capacity": heat_capacity,
                "helmholts_free_energy": helmholts_free_energy,
            },
        ).write_csv(
            "rawphonon.csv",
        )

    import matplotlib.pyplot as plt

    plt.plot(temp, entropy, c="red", label="Entropy (J/K/mol)")
    plt.plot(temp, heat_capacity, c="green", label="Heat Capacity (J/K/mol)")
    plt.plot(
        temp,
        helmholts_free_energy,
        c="blue",
        label="Helmholtz Free Energy (kJ/mol)",
    )
    plt.xlabel("Temperature(K)")
    plt.ylabel("Thermal Properties")
    plt.tick_params(direction="in")  # Tick marks are inside
    plt.grid(alpha=0.2)
    plt.legend()
    plt.title("Thermal")

    plt.tight_layout()
    # save and show
    if figname:
        import os

        absfig = os.path.abspath(figname)
        os.makedirs(os.path.dirname(absfig), exist_ok=True)
        plt.savefig(absfig, dpi=300)
        logger.info(f"==> {absfig}")
    if show:
        plt.show()


def plot_polarization_figure(
    directory: str,
    repetition: int = 2,
    annotation: bool = False,
    annotation_style: int = 1,
    show: bool = True,
    figname: str = "pol.png",
    raw: bool = False,
):
    """Plot the polarization results of the iron electrode

    Parameters
    ----------
    directory:
        Main directory for the iron polarization calculation task
    repetition:
        Number of times to repeat drawing along the upper (or lower) direction, default 2
    annotation:
        Whether to display the polarization values of the iron electrodes at the beginning and end configurations, displayed by default
    show:
        Interactive display of the image, default True
    figname:
        Image save path, default 'pol.png'
    raw:
        Whether to save the raw data to a CSV file

    Returns
    -------
    axes: matplotlib.axes._subplots.AxesSubplot
        Can be passed to other functions for further processing

    Examples
    --------
    >>> from dspawpy.plot import plot_polarization_figure
    >>> result = plot_polarization_figure(directory='tests/2.20', figname='tests/outputs/doctest/pol1.png', show=False, annotation=True, annotation_style=1)
    >>> result = plot_polarization_figure(directory='tests/2.20', figname='tests/outputs/doctest/pol2.png', show=False, annotation=True, annotation_style=2)

    """
    assert repetition >= 0, "The number of repetitions must be a natural number"
    subfolders, quantum, totals = _get_subfolders_quantum_totals(directory)
    number_sfs = [int(sf) for sf in subfolders]
    import matplotlib.pyplot as plt

    fig, axes = plt.subplots(1, 3, sharey=True)
    xyz = ["x", "y", "z"]
    import numpy as np

    for j in range(3):  # x, y, z
        ys = np.empty(shape=(len(subfolders), repetition * 2 + 1))
        for r in range(repetition + 1):
            ys[:, repetition - r] = totals[:, j] - quantum[j] * r
            ys[:, repetition + r] = totals[:, j] + quantum[j] * r

        axes[j].plot(number_sfs, ys, ".")  # plot
        axes[j].set_title("P%s" % xyz[j])
        axes[j].xaxis.set_ticks(number_sfs)  # Set x-axis ticks
        axes[j].set_xticklabels(labels=subfolders, rotation=90)
        axes[j].grid(axis="x", color="gray", linestyle=":", linewidth=0.5)
        axes[j].tick_params(direction="in")
        # set y ticks using the first and last values
        if annotation:
            if annotation_style == 2:
                style = "arc,angleA=-0,angleB=0,armA=-10,armB=0,rad=0"
                for i in range(repetition * 2 + 1):
                    axes[j].annotate(
                        "%.2f" % ys[0, i],
                        xy=(number_sfs[0], ys[0, i]),
                        xycoords="data",
                        xytext=(number_sfs[-1] + 2, ys[0, i] - 8),
                        textcoords="data",
                        arrowprops=dict(
                            arrowstyle="->",
                            color="black",
                            linewidth=0.75,
                            shrinkA=2,
                            shrinkB=1,
                            connectionstyle=style,
                        ),
                    )
                    axes[j].annotate(
                        "%.2f" % ys[-1, i],
                        xy=(number_sfs[-1], ys[-1, i]),
                        xycoords="data",
                        xytext=(number_sfs[-1] + 2, ys[-1, i] + 8),
                        textcoords="data",
                        arrowprops=dict(
                            arrowstyle="->",
                            color="black",
                            linewidth=0.75,
                            shrinkA=2,
                            shrinkB=1,
                            connectionstyle=style,
                        ),
                    )
            elif annotation_style == 1:
                for i in range(repetition * 2 + 1):
                    axes[j].annotate(
                        text="%.2f" % ys[0, i],
                        xy=(0, ys[0, i]),
                        xytext=(0, ys[0, i] - np.max(ys) / repetition / 5),
                    )
                    axes[j].annotate(
                        text="%.2f" % ys[-1, i],
                        xy=(len(subfolders) - 1, ys[-1, i]),
                        xytext=(
                            len(subfolders) - 1,
                            ys[-1, i] - np.max(ys) / repetition / 5,
                        ),
                    )
            else:
                raise ValueError("annotation_style must be 1 or 2")

        if raw:
            import polars as pl

            pl.DataFrame(ys).write_csv(f"pol_{xyz[j]}.csv")

    plt.tight_layout()
    # save and show
    if figname:
        import os

        absfig = os.path.abspath(figname)
        os.makedirs(os.path.dirname(absfig), exist_ok=True)
        plt.savefig(absfig, dpi=300)
        logger.info(f"==> {absfig}")
    if show:
        plt.show()

    return axes


def _get_subfolders_quantum_totals(directory: str):
    """Return the subdirectories, quantum numbers, and polarization totals for iron polarization calculation tasks;

    Do not create other subdirectories, as they will be incorrectly read

    Parameters
    ----------
    directory：str
        Main directory for iron polarization calculation tasks

    Returns
    -------
    subfolders:
        List of subdirectories
    quantum : np.ndarray
        Quantum numbers, xyz directions, shape=(1, 3)
    totals : np.ndarray
        Total polarization, xyz directions, shape=(len(subfolders), 3)

    """
    import os

    absdir = os.path.abspath(directory)
    raw_subfolders = next(os.walk(absdir))[1]
    subfolders = []
    for subfolder in raw_subfolders:
        assert 0 <= int(subfolder) < 100, (
            f"--> You should rename subfolders to 0~99, but {subfolder} found"
        )
        try:
            assert 0 <= int(subfolder) < 100
            subfolders.append(subfolder)
        except Exception:
            pass
    subfolders.sort()  # Sort in ascending order

    # quantum number if constant across the whole calculation, read only once
    absh5 = f"{os.path.join(absdir, subfolders[0])}/scf.h5"
    absjs = f"{os.path.join(absdir, subfolders[0])}/polarization.json"
    import numpy as np
    from h5py import File

    if os.path.isfile(absjs):
        quantum = np.asarray(File(absh5).get("/PolarizationInfo/Quantum"))
    elif os.path.isfile(absjs):
        with open(absjs) as f:
            from json import load

            quantum = np.asarray(load(f)["PolarizationInfo"]["Quantum"])
    else:
        raise FileNotFoundError(f"No {absh5}/{absjs}")

    totals = np.empty(shape=(len(subfolders), 3))
    # the Total number is not constant, read for each subfolder
    for i, fd in enumerate(subfolders):
        absh5 = f"{os.path.join(absdir, fd)}/scf.h5"
        absjs = f"{os.path.join(absdir, fd)}/polarization.json"
        if os.path.isfile(absh5):
            data = File(f"{os.path.join(absdir, fd)}/scf.h5")
            total = np.asarray(data.get("/PolarizationInfo/Total"))
        elif os.path.isfile(absjs):
            with open(absjs) as f:
                from json import load

                data = load(f)
            total = np.asarray(data["PolarizationInfo"]["Total"], dtype=float)
        else:
            raise FileNotFoundError(f"No {absh5}/{absjs}")
        totals[i] = total

    return subfolders, quantum, totals


def getEwtData(nk, nb, celtot, proj_wt, ef, de, dele):
    import numpy as np

    emin = np.min(celtot) - de
    emax = np.max(celtot) - de

    emin = np.floor(emin - 0.2)
    emax = max(np.ceil(emax) * 1.0, 5.0)

    nps = int((emax - emin) / de)

    X = np.zeros((nps + 1, nk))
    Y = np.zeros((nps + 1, nk))

    X2 = []
    Y2 = []
    Z2 = []

    for ik in range(nk):
        for ip in range(nps + 1):
            omega = ip * de + emin + ef
            X[ip][ik] = ik
            Y[ip][ik] = ip * de + emin
            ewts_value = 0
            for ib in range(nb):
                smearing = dele / np.pi / ((omega - celtot[ib][ik]) ** 2 + dele**2)
                ewts_value += smearing * proj_wt[ib][ik]
            if ewts_value > 0.01:
                X2.append(ik)
                Y2.append(ip * de + emin)
                Z2.append(ewts_value)

    Z2_half = max(Z2) / 2

    for i, x in enumerate(Z2):
        if x > Z2_half:
            Z2[i] = Z2_half

    return X2, Y2, Z2, emin


def _read_aimd_converge_data(
    datafile: str,
    index: Optional[str] = None,
    verbose: bool = False,
):
    """Read data specified by index from the path in datafile, and return two arrays xs and ys for plotting.

    Parameters
    ----------
    datafile:
        Path to the HDF5 file, e.g., 'aimd.h5' or ['aimd.h5', 'aimd2.h5']
    index:
        ID, default None

    Returns
    -------
    xs : np.ndarray
        x-axis data
    ys : np.ndarray
        y-axis data

    """
    import numpy as np

    if isinstance(datafile, list):
        xs = []
        ys = []
        for i, df in enumerate(datafile):
            # concentrate returned np.ndarray
            x, y = _read_aimd_converge_data(df, index)
            xs.extend(x)
            ys.extend(y)
        xs = np.linspace(1, len(xs), len(xs))
        return xs, ys

    # search datafile in the given directory
    elif isinstance(datafile, str):
        from dspawpy.io.utils import get_absfile

        absfile = get_absfile(datafile, task="aimd", verbose=verbose)

        if datafile.endswith("h5"):
            from h5py import File

            hf = File(absfile)  # Load the h5 file
            Nstep = (
                len(np.asarray(hf.get("/Structures"))) - 2
            )  # Number of steps (may include unfinished ones)
            ys = np.full(Nstep, np.nan)  # Prepare an empty array
            # Starting to read
            if index == "5":
                for i in range(1, Nstep + 1):
                    ys[i - 1] = np.linalg.det(
                        np.asarray(hf.get("/Structures/Step-%d/Lattice" % i)),
                    )
            else:
                map = {
                    "1": "IonsKineticEnergy",
                    "2": "TotalEnergy0",
                    "3": "PressureKinetic",
                    "4": "Temperature",
                }
                if index == "3" and "PressureKinetic" not in np.asarray(
                    hf.get("/AimdInfo/Step-1"),
                ):
                    logger.warning(
                        "Ensemble is neither NPT nor NPH, no PressureKinetic found for subfigure 3!",
                    )
                else:
                    for i in range(1, Nstep + 1):
                        assert index is not None
                        try:
                            ys[i - 1] = np.asarray(
                                hf.get("/AimdInfo/Step-%d/%s" % (i, map[index])),
                            )[0]
                        except Exception:
                            ys[i - 1] = 0
                            ys = np.delete(ys, -1)
                            logger.warning(
                                "-> AIMD task stopped at Nstep=%s, failed to read its %s value"
                                % (Nstep, map[index]),
                            )
                            break

            Nstep = len(ys)  # Steps updated to the actual number of completed steps
        elif datafile.endswith("json"):
            from json import load

            with open(absfile) as f:
                data = load(f)
            Nstep = len(data["Structures"])
            ys = np.zeros(Nstep)

            if index == "5":
                for i in range(Nstep):
                    ys[i] = np.linalg.det(
                        np.asarray(data["Structures"][i]["Lattice"]).reshape(3, 3),
                    )
            else:
                map = {
                    "1": "IonsKineticEnergy",
                    "2": "TotalEnergy0",
                    "3": "PressureKinetic",
                    "4": "Temperature",
                }
                assert index is not None
                if index in ["1", "2"]:
                    ys = data["AimdInfo"]["Energy"][map[index]]

                elif index == "3":
                    raise ValueError(
                        "PressureKinetic array is not written in json file, so dspawapy can not read it!",
                    )
                else:
                    ys = data["AimdInfo"][map[index]]

            Nstep = len(ys)  # Steps updated to the actual number of completed steps

        # Return two arrays, xs and ys
        return np.linspace(1, Nstep, Nstep), np.asarray(ys)  # type: ignore

    else:
        raise TypeError("datafile must be str or list")


def pltbd(
    bdp: "BSDOSPlotter",
    bs: "BandStructureSymmLine",
    dos: Optional["CompleteDos"] = None,
    demax: float = 0.1,
    ylim: Optional[List[float]] = None,
    alpha: float = 0.3,
    colors: Optional[List[str]] = None,
    filename: str = "banddos.png",
    dpi: int = 300,
):
    """Optimized from get_plot() of BSDOSPlotter class, can draw (projected) band structure and density of states plots.

    Improvements:

    - DOS uses semi-transparent shading
    - Fermi level is represented by a dashed line, with the left and right Fermi levels kept at the same height
    - Line color, width, and type are customizable
    - Allows setting the band range
    - If there is no magnetism, the band legend is not necessary to display
    - Band structures near the Fermi level can be highlighted
    - Can use LaTeX fonts
    - Reduced spacing between left and right plots

    This function has some improvements over the original but still has shortcomings, and more options may be opened in the future. Please pay attention to parameter settings when using it.

    Parameters
    ----------
    bdp:
        `BSDOSPlotter class object`
    bs:
        Band structure data
    dos:
        Density of states data
    demax:
        The bands within the demax range near the Fermi level will be highlighted
    ylim:
        Y-axis limits
    alpha:
        Transparency of the density plot
    colors:
        Color list
    filename:
        Saved filename
    dpi:
        Image DPI

    Examples
    --------
    >>> from dspawpy.io.read import get_band_data, get_dos_data
    >>> from pymatgen.electronic_structure.plotter import BSDOSPlotter
    >>> from dspawpy.plot import pltbd
    >>> bdp = BSDOSPlotter(
    ...     bs_projection=None,  # Band structure projection method, None means no projection
    ...     dos_projection="elements")  # Projection method for DOS, None means no projection
    >>> band_data = get_band_data(
    ...     band_dir='tests/supplement/banddos/band.json',  # Band data
    ...     zero_to_efermi=True,  # For non-metallic systems, the zero point energy should be shifted to the Fermi level,
    ... )
    >>> dos_data = get_dos_data(dos_dir='tests/supplement/banddos/dos.json')
    >>> pltbd(bdp, band_data, dos_data, ylim=(-2,4), filename='tests/outputs/doctest/newbd.png') # doctest: +ELLIPSIS
    <module 'matplotlib.pyplot' from '...pyplot.py'>

    Compared to the previous version:

    >>> bdp = BSDOSPlotter(bs_projection=None,dos_projection="elements")
    >>> plt = bdp.get_plot(bs=band_data, dos=dos_data)

    for old version pymatgen, it returns plt, otherwise may return axes, ref to userscripts for how to user them.

    """
    import numpy as np
    import palettable

    if colors is None:
        _colors = palettable.colorbrewer.qualitative.Set1_9.mpl_colors  # type: ignore
    else:
        _colors = colors

    # self -> BSDOSPlotter
    bs_projection = bdp.bs_projection
    dos_projection = bdp.dos_projection
    vb_energy_range = bdp.vb_energy_range
    cb_energy_range = bdp.cb_energy_range
    fixed_cb_energy = bdp.fixed_cb_energy
    egrid_interval = bdp.egrid_interval
    axis_fontsize = bdp.axis_fontsize
    tick_fontsize = bdp.tick_fontsize
    legend_fontsize = bdp.legend_fontsize
    bs_legend = bdp.bs_legend
    dos_legend = bdp.dos_legend
    rgb_legend = bdp.rgb_legend
    fig_size = bdp.fig_size

    if dos:
        elements = [e.symbol for e in dos.structure.composition.elements]
    elif bs_projection and bs.structure:
        elements = [e.symbol for e in bs.structure.composition.elements]
    else:
        elements = []

    rgb_legend = (
        rgb_legend
        and bs_projection
        and bs_projection.lower() == "elements"
        and len(elements) in [2, 3, 4]
    )

    if (
        bs_projection
        and bs_projection.lower() == "elements"
        and (len(elements) not in [2, 3, 4] or not bs.get_projection_on_elements())
    ):
        logger.warning(
            "Cannot get element projected data; either the projection data "
            "doesn't exist, or you don't have a compound with exactly 2 "
            "or 3 or 4 unique elements.",
        )
        bs_projection = None

    # specify energy range of plot
    emin = -vb_energy_range
    emax = (
        cb_energy_range
        if fixed_cb_energy
        else cb_energy_range + bs.get_band_gap()["energy"]
    )

    # initialize all the k-point labels and k-point x-distances for bs plot
    xlabels = []  # all symmetry point labels on x-axis
    xlabel_distances = []  # positions of symmetry point x-labels

    x_distances_list = []
    prev_right_klabel = (
        None  # used to determine which branches require a midline separator
    )

    for branch in bs.branches:
        x_distances = []

        # get left and right kpoint labels of this branch
        left_k, right_k = branch["name"].split("-")

        # add $ notation for LaTeX kpoint labels
        if left_k[0] == "\\" or "_" in left_k:
            left_k = "$" + left_k + "$"
        if right_k[0] == "\\" or "_" in right_k:
            right_k = "$" + right_k + "$"

        # add left k label to list of labels
        if prev_right_klabel is None:
            xlabels.append(left_k)
            xlabel_distances.append(0)
        elif prev_right_klabel != left_k:  # used for pipe separator
            xlabels[-1] = xlabels[-1] + "$\\mid$ " + left_k

        # add right k label to list of labels
        xlabels.append(right_k)
        prev_right_klabel = right_k

        # add x-coordinates for labels
        left_kpoint = bs.kpoints[branch["start_index"]].cart_coords
        right_kpoint = bs.kpoints[branch["end_index"]].cart_coords
        distance = np.linalg.norm(right_kpoint - left_kpoint)
        xlabel_distances.append(xlabel_distances[-1] + distance)  # type: ignore

        # add x-coordinates for kpoint data
        npts = branch["end_index"] - branch["start_index"]
        distance_interval = distance / npts
        x_distances.append(xlabel_distances[-2])
        for _ in range(npts):
            x_distances.append(x_distances[-1] + distance_interval)
        x_distances_list.append(x_distances)

    # set up bs and dos plot
    from matplotlib.gridspec import GridSpec

    gs = GridSpec(1, 2, width_ratios=[2, 1], wspace=0) if dos else GridSpec(1, 1)

    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=fig_size)
    fig.patch.set_facecolor("white")
    bs_ax = plt.subplot(gs[0])
    # set basic axes limits for the plot
    bs_ax.set_xlim(0, x_distances_list[-1][-1])
    emin = ylim[0] if ylim else emin
    emax = ylim[1] if ylim else emax
    bs_ax.set_ylim(emin, emax)
    # add BS xticks, labels, etc.
    bs_ax.set_xticks(xlabel_distances)
    bs_ax.set_xticklabels(xlabels, size=tick_fontsize)
    bs_ax.set_xlabel("Wavevector $k$", fontsize=axis_fontsize)
    bs_ax.set_ylabel("$E-E_F$ / eV", fontsize=axis_fontsize)

    # add BS fermi level line at E=0 and gridlines
    bs_ax.hlines(y=0, xmin=0, xmax=x_distances_list[-1][-1], color="k", lw=1, ls="--")
    bs_ax.set_yticks(np.arange(emin, emax + 1e-5, egrid_interval))
    bs_ax.set_yticklabels(
        np.arange(emin, emax + 1e-5, egrid_interval),
        size=tick_fontsize,
    )
    bs_ax.set_axisbelow(True)
    bs_ax.grid(color=[0.5, 0.5, 0.5], linestyle="dotted", linewidth=1)
    # renormalize the band energy to the Fermi level
    band_energies = {}
    spin_count = 0
    from pymatgen.electronic_structure.core import Spin

    for spin in (Spin.up, Spin.down):
        if spin in bs.bands:
            spin_count += 1
            band_energies[spin] = []
            for band in bs.bands[spin]:
                band = cast(List[float], band)
                band_energies[spin].append([e - bs.efermi for e in band])  # type: ignore

    # plot the colored band structure lines
    for spin in (Spin.up, Spin.down):
        if spin in band_energies:
            for band_idx, band in enumerate(band_energies[spin]):
                current_pos = 0
                for x_distances in x_distances_list:
                    sub_band = band[current_pos : current_pos + len(x_distances)]
                    # ! highlight the band close to the Fermi level
                    if np.max(np.abs(sub_band)) < demax:
                        bs_ax.plot(x_distances, sub_band, color=_colors[0], linewidth=3)
                    elif spin == Spin.up:
                        bs_ax.plot(
                            x_distances,
                            sub_band,
                            color=_colors[0],
                            linewidth=1,
                        )
                    elif spin == Spin.down:
                        bs_ax.plot(
                            x_distances,
                            sub_band,
                            color=_colors[1],
                            linewidth=1,
                        )
                    current_pos += len(x_distances)

    # add legend for band structure
    if spin_count == 2:
        if bs_legend and not rgb_legend:
            handles = []

            if bs_projection is None:
                import matplotlib.lines as mlines

                handles = [
                    mlines.Line2D(
                        [],
                        [],
                        linewidth=2,
                        color=_colors[0],
                        label="spin up",
                    ),
                    mlines.Line2D(
                        [],
                        [],
                        linewidth=2,
                        color=_colors[1],
                        label="spin down",
                    ),
                ]

            bs_ax.legend(
                handles=handles,
                fancybox=True,
                prop={"size": legend_fontsize},
                loc=bs_legend,
            )

    # renormalize the DOS energies to Fermi level
    from pymatgen.electronic_structure.core import OrbitalType

    if dos:
        dos_ax = plt.subplot(gs[1])
        dos_ax.set_ylim(emin, emax)
        dos_ax.set_yticks(np.arange(emin, emax + 1e-5, egrid_interval))
        dos_ax.set_yticklabels([])
        dos_ax.grid(color=[0.5, 0.5, 0.5], linestyle="dotted", linewidth=1)

        dos_energies = [e - dos.efermi for e in dos.energies]
        # Plot the DOS and projected DOS
        for spin in (Spin.up, Spin.down):
            if spin in dos.densities:
                # plot the total DOS
                dos_densities = dos.densities[spin] * int(spin)
                label = "total" if spin == Spin.up else None
                dos_ax.plot(dos_densities, dos_energies, color=(0.6, 0.6, 0.6))
                dos_ax.fill_betweenx(
                    dos_energies,
                    0,
                    dos_densities,
                    color=(0.7, 0.7, 0.7),
                    alpha=alpha,
                    # facecolor=(0.7, 0.7, 0.7),
                    label=label,
                )
                if dos_projection is None:
                    pass
                elif dos_projection.lower() == "elements":
                    # plot the atom-projected DOS
                    from pymatgen.core.periodic_table import Element

                    for idx, el in enumerate(elements):
                        if spin_count == 1:
                            el_dos = dos.get_element_dos()
                            dos_densities = el_dos[Element(el)].densities[spin] * int(
                                spin,
                            )
                            label = el if spin == Spin.up else None
                            dos_ax.plot(dos_densities, dos_energies, color=_colors[idx])
                            dos_ax.fill_betweenx(
                                dos_energies,
                                0,
                                dos_densities,
                                color=_colors[idx],
                                alpha=alpha,
                                label=label,
                            )
                        elif spin_count == 2:
                            if el not in ["O", "Cu"]:  # TODO generalize
                                continue
                            el_spd_dos = dos.get_element_spd_dos(el)
                            for idx, orb in enumerate([OrbitalType.p, OrbitalType.d]):
                                if (
                                    orb == OrbitalType.p
                                    and el == "Cu"
                                    or orb == OrbitalType.d
                                    and el == "O"
                                ):
                                    continue
                                if orb in el_spd_dos:
                                    dos_densities = el_spd_dos[orb].densities[
                                        spin
                                    ] * int(spin)
                                    label = f"{el}-{orb}" if spin == Spin.up else None  # type: ignore
                                    dos_ax.plot(
                                        dos_densities,
                                        dos_energies,
                                        color=_colors[idx],
                                    )
                                    dos_ax.fill_betweenx(
                                        dos_energies,
                                        0,
                                        dos_densities,
                                        color=_colors[idx],
                                        alpha=alpha,
                                        label=label,
                                    )

        # get index of lowest and highest energy being plotted, used to help auto-scale DOS x-axis
        emin_idx = next(x[0] for x in enumerate(dos_energies) if x[1] >= emin)
        emax_idx = len(dos_energies) - next(
            x[0] for x in enumerate(reversed(dos_energies)) if x[1] <= emax
        )

        # determine DOS x-axis range
        dos_xmin = (
            0
            if Spin.down not in dos.densities
            else -max(dos.densities[Spin.down][emin_idx : emax_idx + 1] * 1.05)
        )
        dos_xmax = max(
            [max(dos.densities[Spin.up][emin_idx:emax_idx]) * 1.05, abs(dos_xmin)],
        )

        # set up the DOS x-axis and add Fermi level line
        dos_ax.set_xlim(dos_xmin, dos_xmax)
        dos_ax.set_xticklabels([])
        dos_ax.hlines(y=0, xmin=dos_xmin, xmax=dos_xmax, color="k", lw=1, ls="--")
        dos_ax.set_xlabel("DOS", fontsize=axis_fontsize)
        dos_ax.set_ylim(emin, emax)

        # add legend for DOS
        if dos_legend:
            dos_ax.legend(
                fancybox=True,
                prop={"size": legend_fontsize},
                loc=dos_legend,
            )

    plt.subplots_adjust(wspace=0.1)
    import os

    os.makedirs(os.path.dirname(os.path.abspath(filename)), exist_ok=True)
    plt.tight_layout()
    plt.savefig(filename, dpi=dpi)
    # plt.show()

    return plt


def plot_dos(
    dosplotter,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    invert_axes: bool = False,
    beta_dashed: bool = False,
    ax=None,
    raw: bool = False,
):
    """Get a matplotlib plot showing the DOS.

    Parameters
    ----------
    dosplotter (DosPlotter): A DosPlotter object containing a complete Dos.
    xlim (tuple[float, float]): The energy axis limits. Defaults to None for automatic
        determination.
    ylim (tuple[float, float]): The y-axis limits. Defaults to None for automatic determination.
    invert_axes (bool): Whether to invert the x and y axes. Enables chemist style DOS plotting.
        Defaults to False.
    beta_dashed (bool): Plots the beta spin channel with a dashed line. Defaults to False.
    raw (bool): Whether to save raw data to csv file. Defaults to False.

    Returns
    -------
    plt.Axes: matplotlib Axes object.

    Example
    -------
    >>> from dspawpy.io.read import get_dos_data
    >>> from pymatgen.electronic_structure.plotter import DosPlotter
    >>> from dspawpy.plot import plot_dos

    >>> dos_data = get_dos_data(
    ...    dos_dir="tests/3.2.4/dos.h5",
    ...    return_dos=False)
    >>> dos_plotter = DosPlotter(
    ... zero_at_efermi=True,
    ... stack=False,
    ... sigma=None)
    >>> dos_plotter.add_dos(label="total dos", dos=dos_data)
    >>> ax = plot_dos(
    ...  dosplotter=dos_plotter,
    ...  xlim=[-10, 5],
    ...  ylim=[-15, 15])
    >>> fig = ax.get_figure()
    >>> fig.savefig("tests/outputs/doctest/plot_dos.png")

    """
    import matplotlib.pyplot as plt
    import palettable

    if not ax:
        from pymatgen.util.plotting import pretty_plot

        ax = pretty_plot(12, 8)  # may be plt for old pmg

    n_colors = min(9, max(3, len(dosplotter._doses)))

    # https://jiffyclub.github.io/palettable/colorbrewer/qualitative/
    colors = palettable.colorbrewer.qualitative.Set1_9.mpl_colors  # type: ignore

    ys = None
    all_densities = []
    all_energies = []
    if isinstance(ax, type(plt)):
        ax = ax.gca()  # type: ignore

    # Note that this complicated processing of energies is to allow for
    # stacked plots in matplotlib.
    import numpy as np
    from pymatgen.electronic_structure.core import Spin

    for dos in dosplotter._doses.values():
        energies = dos["energies"]
        densities = dos["densities"]
        if not ys:
            ys = {
                Spin.up: np.zeros(energies.shape),
                Spin.down: np.zeros(energies.shape),
            }
        new_dens = {}
        for spin in [Spin.up, Spin.down]:
            if spin in densities:
                if dosplotter.stack:
                    ys[spin] += densities[spin]
                    new_dens[spin] = ys[spin].copy()
                else:
                    new_dens[spin] = densities[spin]
        all_energies.append(energies)
        all_densities.append(new_dens)

    keys = list(reversed(dosplotter._doses))
    all_densities.reverse()
    all_energies.reverse()
    all_pts = []

    dd = {}
    for idx, key in enumerate(keys):
        for spin in [Spin.up, Spin.down]:
            if spin in all_densities[idx]:
                energy = all_energies[idx]
                densities = list(int(spin) * all_densities[idx][spin])
                if invert_axes:
                    x = densities
                    y = energy
                else:
                    x = energy
                    y = densities
                all_pts.extend(list(zip(x, y)))
                if dosplotter.stack:
                    ax.fill(x, y, color=colors[idx % n_colors], label=str(key))
                elif spin == Spin.down and beta_dashed:
                    ax.plot(
                        x,
                        y,
                        color=colors[idx % n_colors],
                        label=str(key),
                        linestyle="--",
                        linewidth=3,
                    )
                else:
                    ax.plot(
                        x,
                        y,
                        color=colors[idx % n_colors],
                        label=str(key),
                        linewidth=3,
                    )

                dd[f"key={key}_spin={spin}_x"] = x
                dd[f"key={key}_spin={spin}_y"] = y

    if xlim:
        ax.set_xlim(xlim)
    if ylim:
        ax.set_ylim(ylim)
    elif not invert_axes:
        xlim = ax.get_xlim()
        relevant_y = [p[1] for p in all_pts if xlim[0] < p[0] < xlim[1]]
        ax.set_ylim((min(relevant_y), max(relevant_y)))
    if not xlim and invert_axes:
        ylim = ax.get_ylim()
        relevant_y = [p[0] for p in all_pts if ylim[0] < p[1] < ylim[1]]
        ax.set_xlim((min(relevant_y), max(relevant_y)))

    if dosplotter.zero_at_efermi:
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        ax.plot(xlim, [0, 0], "k--", linewidth=2) if invert_axes else ax.plot(
            [0, 0],
            ylim,
            "k--",
            linewidth=2,
        )

    if invert_axes:
        ax.set_ylabel("Energies (eV)")
        # ax.set_xlabel(
        #     f"Density of states (states/eV{'/Å³' if hasattr(dosplotter, 'norm_val') else ''})",
        # )
        ax.set_xlabel("Density of states (states/eV)")
        ax.axvline(x=0, color="k", linestyle="--", linewidth=2)
    else:
        ax.set_xlabel("Energies (eV)")
        ax.set_ylabel("Density of states (states/eV)")
        # if hasattr(dosplotter, "_norm_val"):
        #     # if dosplotter._norm_val:
        #     ax.set_ylabel("Density of states (states/eV/Å³)")
        # else:
        #     ax.set_ylabel("Density of states (states/eV)")
        ax.axhline(y=0, color="k", linestyle="--", linewidth=2)

    # Remove duplicate labels with a dictionary
    handles, labels = ax.get_legend_handles_labels()
    label_dict = dict(zip(labels, handles))
    ax.legend(label_dict.values(), label_dict)
    legend_text = (
        ax.get_legend().get_texts()
    )  # all the text.Text instance in the legend
    plt.setp(legend_text, fontsize=30)
    plt.tight_layout()

    if raw:
        import polars as pl

        df = pl.DataFrame(dd)
        df.write_csv("dos_raw.csv")
    return ax



import os
from typing import Literal

from loguru import logger
from prompt_toolkit import prompt

from dspawpy import auto_test_cli
from dspawpy.cli import list_ds, pc_h5
from dspawpy.cli.menu_prompts import Dio, Dresponse, Dselect


def s15_1(language: Literal["EN", "CN"]) -> dict:
    """View the HDF5 data structure.

    h5glance name.h5
    """
    if auto_test_cli:
        for D in list_ds:
            if D["menu"] == 151:
                os.system(f"h5glance {D['h5file']}")
        D = {}
    else:
        D = {}
        D["menu"] = 151
        D["h5file"] = prompt(Dio[language]["h5file"], completer=pc_h5)
        os.system(f"h5glance {D['h5file']}")
    return D


def s15_2(language: Literal["EN", "CN"]) -> dict:
    """View the contents of HDF5 data.

    Supports viewing data for a single key, and Unix systems also support viewing data for all keys (using less for pagination)
    Windows systems do not have less; they will print out all the content directly.
    (Suggest using the HDFViewer window program instead of dspawpy)

    h5glance name.h5 -
    ptdump -d name.h5
    """
    if auto_test_cli:
        from platform import system as cli_system

        for D in list_ds:
            if D["menu"] == 152:
                if D["is_all"] == "y":
                    if cli_system() == "Windows":
                        cmd = f"ptdump -d {D['h5file']}"
                    else:
                        cmd = f"ptdump -d {D['h5file']} | less"
                    os.system(cmd)
                else:
                    cmd = f"h5glance {D['h5file']} -"
                    os.system(cmd)
        D = {}
    else:
        D = {}
        D["menu"] = 152
        D["h5file"] = prompt(Dio[language]["h5file"], completer=pc_h5)
        D["is_all"] = prompt(Dselect[language][18])
        if D["is_all"] == "y":
            from platform import system

            # requires PyTables, `pip install tables`
            if system() == "Windows":
                cmd = f"ptdump -d {D['h5file']}"
            else:
                cmd = f"ptdump -d {D['h5file']} | less"

            os.system(cmd)
        else:
            while True:
                logger.info(Dresponse[language][15])
                cmd = f"h5glance {D['h5file']} -"
                os.system(cmd)
                if prompt(Dselect[language][19]) != "y":
                    break
    return D



import threading
from typing import Literal

from loguru import logger
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter

from dspawpy import auto_test_cli
from dspawpy.cli import list_ds, pc
from dspawpy.cli.auxiliary import get_input, get_lims
from dspawpy.cli.menu_prompts import Dio, Dparameter, Dresponse, Dselect


def s6_1(language: Literal["EN", "CN"]) -> dict:
    """Display band structure and density of states on one plot"""
    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import (
            BSDOSPlotter as cli_BSDOSPlotter,
        )

        from dspawpy.io.read import (
            get_band_data as cli_get_band_data,
        )
        from dspawpy.io.read import (
            get_dos_data as cli_get_dos_data,
        )
        from dspawpy.plot import pltbd as cli_pltbd

        D = {}
        for D in list_ds:
            if D["menu"] == 61:
                band_data = cli_get_band_data(D["bandf"])
                dos_data = cli_get_dos_data(D["dosf"])

                bdp = cli_BSDOSPlotter(bs_projection=None, dos_projection=None)  # pyright: ignore [reportArgumentType]

                cli_pltbd(
                    bdp,
                    band_data,
                    dos_data,  # type: ignore
                    ylim=D["ylims"],
                    filename=D["outfile"],
                )

    else:

        def imp():
            global BSDOSPlotter, pltbd, get_band_data, get_dos_data
            from pymatgen.electronic_structure.plotter import BSDOSPlotter

            from dspawpy.io.read import get_band_data, get_dos_data
            from dspawpy.plot import pltbd

        import_thread = threading.Thread(target=imp)
        import_thread.start()

        D = {}
        D["menu"] = 61
        D["bandf"] = prompt(Dio[language]["band"], completer=pc)
        D["dosf"] = prompt(Dio[language]["dos"], completer=pc)
        D["ylims"] = get_lims(Dparameter[language][10], language)
        D["outfile"] = prompt(Dio[language]["figure"], completer=pc)
        import_thread.join()
        logger.info(Dresponse[language][16])

        band_data = get_band_data(D["bandf"])
        dos_data = get_dos_data(D["dosf"])

        bdp = BSDOSPlotter(bs_projection=None, dos_projection=None)  # pyright: ignore [reportArgumentType]

        pltbd(bdp, band_data, dos_data, ylim=D["ylims"], filename=D["outfile"])  # type: ignore

    return D


def s6_2(language: Literal["EN", "CN"]) -> dict:
    "Display band structure and projected density of states on one plot"
    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import (
            BSDOSPlotter as cli_BSDOSPlotter,
        )

        from dspawpy.io.read import (
            get_band_data as cli_get_band_data,
        )
        from dspawpy.io.read import (
            get_dos_data as cli_get_dos_data,
        )
        from dspawpy.plot import pltbd as cli_pltbd

        for D in list_ds:
            if D["menu"] == 62:
                band_data = cli_get_band_data(D["bandf"])
                dos_data = cli_get_dos_data(D["dosf"])

                bdp = cli_BSDOSPlotter(
                    bs_projection=D["pband"], dos_projection=D["pdos"]
                )  # type: ignore

                cli_pltbd(
                    bdp,
                    band_data,
                    dos_data,  # type: ignore
                    ylim=D["ylims"],
                    filename=D["outfile"],
                )

        D = {}
    else:

        def imp():
            global BSDOSPlotter, pltbd, get_band_data, get_dos_data
            from pymatgen.electronic_structure.plotter import BSDOSPlotter

            from dspawpy.io.read import get_band_data, get_dos_data
            from dspawpy.plot import pltbd

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 61
        D["bandf"] = prompt(Dio[language]["band"], completer=pc)
        D["dosf"] = prompt(Dio[language]["dos"], completer=pc)
        D["ylims"] = get_lims(Dparameter[language][10], language)
        D["outfile"] = prompt(Dio[language]["figure"], completer=pc)
        projs = ["elements"]
        D["pband"] = get_input(
            Dselect[language][17],
            projs,
            completer=WordCompleter(projs),
            allow_empty=True,
        )
        projs2 = ["elements", "orbitals"]
        D["pdos"] = get_input(
            Dselect[language][17],
            projs2,
            completer=WordCompleter(projs2),
            allow_empty=True,
        )
        import_thread.join()
        logger.info(Dresponse[language][16])

        band_data = get_band_data(D["bandf"])
        dos_data = get_dos_data(D["dosf"])

        bdp = BSDOSPlotter(bs_projection=D["pband"], dos_projection=D["pdos"])  # type: ignore

        pltbd(bdp, band_data, dos_data, ylim=D["ylims"], filename=D["outfile"])  # type: ignore

    return D



from typing import TYPE_CHECKING, Optional, Sequence

from loguru import logger

if TYPE_CHECKING:
    from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine


def get_band_data(
    band_dir: str,
    syst_dir: Optional[str] = None,
    efermi: Optional[float] = None,
    zero_to_efermi: bool = False,
    verbose: bool = False,
) -> "BandStructureSymmLine":
    """Reads band structure data from an h5 or json file and constructs a BandStructureSymmLine object.

    Parameters
    ----------
    band_dir
        - Path to the band structure file, band.h5 / band.json, or a directory containing band.h5 / band.json
        - Note that wannier.h5 can also be read using this function, but band_dir does not support folder types
    syst_dir
        Path to system.json, prepared only for auxiliary processing of Wannier data (structure and Fermi level are read from it)
    efermi
        Fermi level, if the Fermi level in the h5 file is incorrect, it can be specified using this parameter
    zero_to_efermi
        Whether to shift the Fermi level to 0

    Returns
    -------
    BandStructureSymmLine

    Examples
    --------
    >>> from dspawpy.io.read import get_band_data
    >>> band = get_band_data(band_dir='tests/2.3/band.h5')
    >>> band = get_band_data(band_dir='tests/2.4/band.h5')
    >>> band = get_band_data(band_dir='tests/2.4/band.json')

    If you want to process Wannier band structures by specifying wannier.json, you need to additionally specify the syst_dir parameter.

    >>> band = get_band_data(band_dir='tests/2.30/wannier.h5')
    >>> band = get_band_data(band_dir='tests/2.30/wannier.json', syst_dir='tests/2.30/system.json')

    """
    if efermi is not None and zero_to_efermi:
        raise ValueError(
            "efermi and zero_to_efermi should not be set at the same time!",
        )

    from dspawpy.io.utils import get_absfile

    absfile = get_absfile(
        band_dir,
        task="band",
        verbose=verbose,
    )  # give wannier.h5 also work, verbose=verboses
    if absfile.endswith(".h5"):
        band = load_h5(absfile)
        from h5py import File

        raw = File(absfile, "r").keys()
        if "/WannBandInfo/NumberOfBand" in raw:
            (
                structure,
                kpoints,
                eigenvals,
                rEf,
                labels_dict,
                projections,
            ) = _get_band_data_h5(band, iwan=True, zero_to_efermi=zero_to_efermi)
        elif "/BandInfo/NumberOfBand" in raw:
            (
                structure,
                kpoints,
                eigenvals,
                rEf,
                labels_dict,
                projections,
            ) = _get_band_data_h5(band, iwan=False, zero_to_efermi=zero_to_efermi)
        else:
            raise KeyError("BandInfo or WannBandInfo key not found in h5file!")
    elif absfile.endswith(".json"):
        with open(absfile) as fin:
            from json import load

            band = load(fin)
        if "WannBandInfo" in band.keys():
            assert syst_dir is not None, (
                "system.json is required for processing wannier band info!"
            )
            with open(syst_dir) as system_json:
                from json import load

                syst = load(system_json)
            (
                structure,
                kpoints,
                eigenvals,
                rEf,
                labels_dict,
                projections,
            ) = _get_band_data_json(
                band,
                syst,
                iwan=True,
                zero_to_efermi=zero_to_efermi,
            )
        elif "BandInfo" in band.keys():
            (
                structure,
                kpoints,
                eigenvals,
                rEf,
                labels_dict,
                projections,
            ) = _get_band_data_json(band, iwan=False, zero_to_efermi=zero_to_efermi)
        else:
            raise ValueError(
                f"BandInfo or WannBandInfo key not found in {absfile} file!",
            )
    else:
        raise TypeError(f"{absfile} must be h5 or json file!")

    if efermi:  # Fermi level read directly from h5 might be incorrect, user needs to specify manually
        rEf = efermi  # This is just a temporary solution

    from pymatgen.core.lattice import Lattice

    lattice_new = Lattice(structure.lattice.reciprocal_lattice.matrix)

    from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine

    return BandStructureSymmLine(
        kpoints=kpoints,
        eigenvals=eigenvals,
        lattice=lattice_new,
        efermi=rEf,
        labels_dict=labels_dict,
        structure=structure,
        projections=projections,
    )


def get_dos_data(
    dos_dir: str,
    return_dos: bool = False,
    verbose: bool = False,
):
    """Read density of states (DOS) data from an h5 or json file, and construct a CompleteDos or DOS object

    Parameters
    ----------
    dos_dir:
        Path to the density of states file, dos.h5 / dos.json, or a folder containing dos.h5 / dos.json
    return_dos : bool, optional
        Whether to return the DOS object. If False, a CompleteDos object is returned uniformly (regardless of whether projection was enabled during calculation)

    Returns
    -------
    CompleteDos or Dos

    Examples
    --------
    >>> from dspawpy.io.read import get_dos_data
    >>> dos = get_dos_data(dos_dir='tests/2.5/dos.h5')
    >>> dos = get_dos_data(dos_dir='tests/2.5/dos.h5', return_dos=True)

    """
    from dspawpy.io.utils import get_absfile

    absfile = get_absfile(dos_dir, task="dos", verbose=verbose)
    if absfile.endswith(".h5"):
        dos = load_h5(absfile)
        if return_dos and not dos["/DosInfo/Project"][0]:
            return _get_total_dos(dos)
        else:
            return _get_complete_dos(dos)

    elif absfile.endswith(".json"):
        with open(absfile) as fin:
            from json import load

            dos = load(fin)
        if return_dos and not dos["DosInfo"]["Project"]:
            return _get_total_dos_json(dos)
        else:
            return _get_complete_dos_json(dos)

    else:
        raise TypeError(f"{absfile} must be h5 or json file!")


def get_ele_from_h5(hpath: str = "aimd.h5") -> list:
    """Read the list of elements from the h5 file;
    Multiple ion steps do not save element information in each ion step's Structure; only the element information of the initial structure can be read.

    Parameters
    ----------
    hpath:
        h5 file path

    Returns
    -------
    ele:
        element list, Natom x 1

    Examples
    --------
    >>> from dspawpy.io.read import get_ele_from_h5
    >>> ele = get_ele_from_h5(hpath='tests/2.18/aimd.h5')
    >>> ele
    ['H', 'H_1', 'O']

    """
    import os

    absh5 = os.path.abspath(hpath)
    from h5py import File

    data = File(absh5)
    import numpy as np

    Elements_bytes = np.asarray(data.get("/AtomInfo/Elements"))
    tempdata = np.asarray([i.decode() for i in Elements_bytes])
    ele = "".join(tempdata).split(";")

    return ele


def get_lines_without_comment(filename: str, comment: str = "#") -> list:
    """Read the content of an as file, remove comments, and return a list of lines

    Examples
    --------
    >>> from dspawpy.io.read import get_lines_without_comment
    >>> lines = get_lines_without_comment(filename='tests/2.15/01/structure01.as', comment='#')
    >>> lines
    ['Total number of atoms', '13', 'Lattice', '5.60580000 0.00000000 0.00000000', '0.00000000 5.60580000 0.00000000', '0.00000000 0.00000000 16.81740000', 'Cartesian', 'H 2.48700709 3.85367720 6.93461994', 'Pt 1.40145000 1.40145000 1.98192999', 'Pt 4.20434996 1.40145000 1.98192999', 'Pt 1.40145000 4.20434996 1.98192999', 'Pt 4.20434996 4.20434996 1.98192999', 'Pt 0.00843706 0.00042409 3.91500875', 'Pt 0.00881029 2.80247953 3.91551673', 'Pt 2.81216310 -0.00105882 3.91807627', 'Pt 2.81156629 2.80392163 3.91572506', 'Pt 1.41398585 1.39603492 5.85554462', 'Pt 4.22886663 1.39820574 5.84677553', 'Pt 1.40485707 4.20963461 5.89521929', 'Pt 4.23788559 4.20753128 5.88625580']

    """
    lines = []
    import os

    absfile = os.path.abspath(filename)
    import re

    with open(absfile) as file:
        while True:
            line = file.readline()
            if line:
                line = re.sub(comment + r".*$", "", line)  # remove comment
                line = line.strip()
                if line:
                    lines.append(line)
            else:
                break

    return lines


def get_phonon_band_data(
    phonon_band_dir: str,
    verbose: bool = False,
):
    """Reads phonon band data from an h5 or json file and constructs a PhononBandStructureSymmLine object

    Parameters
    ----------
    phonon_band_dir:
        Path to the band structure file, phonon.h5 / phonon.json, or a folder containing these files

    Returns
    -------
    PhononBandStructureSymmLine

    Examples
    --------
    >>> from dspawpy.io.read import get_phonon_band_data
    >>> band_data = get_phonon_band_data("tests/2.16/phonon.h5") # Read phonon band data
    >>> band_data = get_phonon_band_data("tests/2.16/phonon.json") # Read phonon band data

    """
    from dspawpy.io.utils import get_absfile

    absfile = get_absfile(phonon_band_dir, task="phonon", verbose=verbose)

    if absfile.endswith(".h5"):
        band = load_h5(absfile)
        (
            symmmetry_kpoints,
            symmetry_kPoints_index,
            qpoints,
            structure,
            frequencies,
        ) = _get_phonon_band_data_h5(band)
    elif absfile.endswith(".json"):
        with open(absfile) as fin:
            from json import load

            band = load(fin)
        (
            symmmetry_kpoints,
            symmetry_kPoints_index,
            qpoints,
            structure,
            frequencies,
        ) = _get_phonon_band_data_json(band)
    else:
        raise TypeError(f"{absfile} must be h5 or json file")

    labels_dict = {}
    for i, s in enumerate(symmmetry_kpoints):
        labels_dict[s] = qpoints[symmetry_kPoints_index[i] - 1]
    from pymatgen.core.lattice import Lattice

    lattice_new = Lattice(structure.lattice.reciprocal_lattice.matrix)

    from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine

    return PhononBandStructureSymmLine(
        qpoints=qpoints,  # type: ignore
        frequencies=frequencies,
        lattice=lattice_new,
        has_nac=False,
        labels_dict=labels_dict,
        structure=structure,
    )


def get_phonon_dos_data(
    phonon_dos_dir: str,
    verbose: bool = False,
):
    """Reads phonon density of states data from an h5 or json file, constructs a PhononDos object

    Parameters
    ----------
    phonon_dos_dir:
        Path to the phonon DOS file, phonon_dos.h5 / phonon_dos.json, or a folder containing these files

    Returns
    -------
    PhononDos

    Examples
    --------
    >>> from dspawpy.io.read import get_phonon_dos_data
    >>> phdos = get_phonon_dos_data(phonon_dos_dir='tests/2.16.1/phonon.json')
    >>> phdos = get_phonon_dos_data(phonon_dos_dir='tests/2.16.1/phonon.h5')
    >>> phdos.frequencies
    array([ 0. ,  0.1,  0.2,  0.3,  0.4,  0.5,  0.6,  0.7,  0.8,  0.9,  1. ,
            1.1,  1.2,  1.3,  1.4,  1.5,  1.6,  1.7,  1.8,  1.9,  2. ,  2.1,
            2.2,  2.3,  2.4,  2.5,  2.6,  2.7,  2.8,  2.9,  3. ,  3.1,  3.2,
            3.3,  3.4,  3.5,  3.6,  3.7,  3.8,  3.9,  4. ,  4.1,  4.2,  4.3,
            4.4,  4.5,  4.6,  4.7,  4.8,  4.9,  5. ,  5.1,  5.2,  5.3,  5.4,
            5.5,  5.6,  5.7,  5.8,  5.9,  6. ,  6.1,  6.2,  6.3,  6.4,  6.5,
            6.6,  6.7,  6.8,  6.9,  7. ,  7.1,  7.2,  7.3,  7.4,  7.5,  7.6,
            7.7,  7.8,  7.9,  8. ,  8.1,  8.2,  8.3,  8.4,  8.5,  8.6,  8.7,
            8.8,  8.9,  9. ,  9.1,  9.2,  9.3,  9.4,  9.5,  9.6,  9.7,  9.8,
            9.9, 10. , 10.1, 10.2, 10.3, 10.4, 10.5, 10.6, 10.7, 10.8, 10.9,
           11. , 11.1, 11.2, 11.3, 11.4, 11.5, 11.6, 11.7, 11.8, 11.9, 12. ,
           12.1, 12.2, 12.3, 12.4, 12.5, 12.6, 12.7, 12.8, 12.9, 13. , 13.1,
           13.2, 13.3, 13.4, 13.5, 13.6, 13.7, 13.8, 13.9, 14. , 14.1, 14.2,
           14.3, 14.4, 14.5, 14.6, 14.7, 14.8, 14.9, 15. , 15.1, 15.2, 15.3,
           15.4, 15.5, 15.6, 15.7, 15.8, 15.9, 16. , 16.1, 16.2, 16.3, 16.4,
           16.5, 16.6, 16.7, 16.8, 16.9, 17. , 17.1, 17.2, 17.3, 17.4, 17.5,
           17.6, 17.7, 17.8, 17.9, 18. , 18.1, 18.2, 18.3, 18.4, 18.5, 18.6,
           18.7, 18.8, 18.9, 19. , 19.1, 19.2, 19.3, 19.4, 19.5, 19.6, 19.7,
           19.8, 19.9, 20. ])

    """
    from dspawpy.io.utils import get_absfile

    absfile = get_absfile(phonon_dos_dir, task="phonon_dos", verbose=verbose)
    if absfile.endswith(".h5"):
        dos = load_h5(absfile)
        frequencies = dos["/DosInfo/DosEnergy"]
        densities = dos["/DosInfo/Spin1/Dos"]
    elif absfile.endswith(".json"):
        with open(absfile) as fin:
            from json import load

            dos = load(fin)
        frequencies = dos["DosInfo"]["DosEnergy"]
        densities = dos["DosInfo"]["Spin1"]["Dos"]
    else:
        raise TypeError(f"{absfile} must be h5 or json file")

    from pymatgen.phonon.dos import PhononDos

    return PhononDos(frequencies, densities)


def get_sinfo(
    datafile: str,
    scaled: bool = False,
    si=None,
    ele=None,
    ai=None,
    verbose: bool = False,
):
    r"""Reads structural information from datafile

    Parameters
    ----------
    datafile:
        Path to h5 / json file
    scaled : bool, optional
        Whether to return score coordinates, default is False
    si : int or list or str, optional
        Step in the movement trajectory, counting from 1!
        If slicing is required, use string notation: '1, 10'
        Default is None, returns all steps
    ele : list, optional
        Element list, Natom x 1
        Default to None, read from h5 file
    ai : int or list or str, optional
        The ion step number within the multi-ion steps, counting from 1
        For slicing, use string notation: '1, 10'
        Default is None, returns all ion steps

    Returns
    -------
    Nstep:
        Total number of ionic steps (number of configurations)
    ele:
        Element list, Natom x 1
    pos : np.ndarray
        Coordinate component array, Nstep x Natom x 3
    latv : np.ndarray
        Lattice vector array, Nstep x 3 x 3
    D_mag_fix:
        Information related to magnetic moments and degrees of freedom

    Examples
    --------
    >>> from dspawpy.io.read import get_sinfo
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.h5', scaled=False, si=None, ele=None, ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.h5', scaled=True, si=[1,10], ele=None, ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.h5', scaled=True, si=2, ele=None, ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.h5', scaled=True, si='1:', ele=None, ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.h5', scaled=True, si=None, ele=['H', 'O'], ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.h5', scaled=True, si=None, ele='H', ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.h5', scaled=False, si=None, ele=None, ai=[1,2])
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.h5', scaled=False, si=None, ele=None, ai=1)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.h5', scaled=False, si=None, ele=None, ai='1:')
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.2/rho.h5', scaled=False)

    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.json', scaled=False, si=None, ele=None, ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.json', scaled=True, si=[1,10], ele=None, ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.json', scaled=True, si=2, ele=None, ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.json', scaled=True, si='1:', ele=None, ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.json', scaled=True, si=None, ele=['H', 'O'], ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.json', scaled=True, si=None, ele='H', ai=None)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.json', scaled=False, si=None, ele=None, ai=[1,2])
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.json', scaled=False, si=None, ele=None, ai=1)
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.18/aimd.json', scaled=False, si=None, ele=None, ai='1:')
    >>> Nstep, elements, pos, latv, D_mag_fix = get_sinfo(datafile='tests/2.2/rho.json', scaled=False)

    This information can be used to further construct a Structure object.
    Specifically, refer to the `dspawpy.io.structure.build_Structures_from_datafile` function

    """
    assert ele is None or ai is None, (
        "Cannot select element and atomic number at the same time"
    )

    from dspawpy.io.utils import get_absfile

    absfile = get_absfile(datafile, task="free", verbose=verbose)
    import numpy as np

    D_mag_fix = {}
    if absfile.endswith(".h5"):
        from h5py import File

        hf = File(absfile)  # Load the h5 file

        # decide task type by check the internal key
        if "/Structures" in hf.keys():  # multi-steps
            Total_step = np.asarray(hf.get("/Structures/FinalStep"))[
                0
            ]  # Total number of steps
            if f"/Structures/Step-{Total_step}" not in hf.keys():
                Total_step -= 1  # The final step may not have been saved yet

            if si is not None:  # Step number
                if isinstance(si, int):  # 1
                    indices = [si]

                elif isinstance(si, list) or isinstance(ai, np.ndarray):  # [1,2,3]
                    indices = si

                elif isinstance(si, str):  # ':', '1:'
                    indices = __parse_indices(si, Total_step)

                else:
                    raise ValueError("si=%s is invalid" % si)

                Nstep = len(indices)
            else:
                Nstep = Total_step
                indices = list(range(1, Nstep + 1))

            # Read the element list, which does not change with the step number and does not "merge similar items"
            Elements = np.asarray(get_ele_from_h5(absfile), dtype=object)

            # Start reading unit cells and atomic positions
            lattices = np.empty((Nstep, 3, 3))  # Nstep x 3 x 3
            location = []
            if ele is not None:  # If the user specifies an element
                if isinstance(ele, str):  # Single element symbol, e.g., 'Fe'
                    ele_list = np.asarray(ele, dtype=object)
                    location = np.where(Elements == ele_list)[0]
                # List of multiple element symbols, such as ['Fe', 'O']
                elif isinstance(ele, (list, np.ndarray)):
                    for e in ele:
                        loc = np.where(Elements == e)[0]
                        location.append(loc)
                    location = np.concatenate(location)
                else:
                    raise TypeError("ele=%s is invalid" % ele)
                elements = Elements[location]

            elif ai is not None:  # If the user specifies the atomic number
                if isinstance(ai, int):  # 1
                    ais = [ai]
                elif isinstance(ai, (list, np.ndarray)):  # [1,2,3]
                    ais = ai
                elif isinstance(ai, str):  # ':', '1:'
                    ais = __parse_indices(ai, len(Elements))
                else:
                    raise ValueError("ai=%s is invalid" % ai)
                ais = [
                    i - 1 for i in ais
                ]  # Python uses 0-based indexing, but users count from 1
                elements = Elements[ais]
                location = ais

            else:  # If none are specified
                elements = Elements
                location = list(range(len(Elements)))

            elements = elements.tolist()  # for pretty output
            Natom = len(elements)
            poses = np.empty(shape=(len(indices), Natom, 3))
            wrapped_poses = np.empty(shape=(len(indices), Natom, 3))
            for i, ind in enumerate(indices):  # Step number
                lats = np.asarray(hf.get("/Structures/Step-" + str(ind) + "/Lattice"))
                lattices[i] = lats
                # [x1,y1,z1,x2,y2,z2,x3,y3,z3], ...
                # During structure optimization, fractional coordinates are always output, regardless of what CoordinateType is specified!
                rawpos = np.asarray(
                    hf.get("/Structures/Step-" + str(ind) + "/Position"),
                )  # (3, Natom)
                pos = rawpos[:, location]
                wrapped_pos = pos - np.floor(pos)  # wrap into [0,1)
                wrapped_pos = wrapped_pos.flatten().reshape(-1, 3)
                wrapped_poses[i] = wrapped_pos

            if "/AtomInfo/Fix" in hf.keys():  # fix atom
                atomFixs_raw = np.asarray(hf.get("/AtomInfo/Fix"))
                atomfix = np.asarray(
                    ["True" if _v else "False" for _v in atomFixs_raw],
                ).reshape(-1, 3)
            else:
                atomfix = np.full(shape=(Natom, 3), fill_value="False")

            try:  # fix lattice
                latticeFixs = (
                    np.asarray(hf.get("/AtomInfo/FixLattice")).astype(bool).flatten()
                )
                assert latticeFixs.shape == (9,)
                latticeFixs = latticeFixs.reshape(
                    9,
                )  # (9,)
            except Exception as e:
                if str(e):  # ignore empty AssertionError()
                    print(e)
                latticeFixs = np.full(shape=(9,), fill_value="False")

            # iNoncollinear = False
            try:  # Spin calculation
                if "/MagInfo/TotalMagOnAtom" in hf.keys():  # collinear
                    mag = np.asarray(hf.get("/MagInfo/TotalMagOnAtom"))  # Natom x 1
                    mags = np.repeat(mag[np.newaxis, :], Nstep, axis=0).tolist()
                    D_mag_fix = {
                        "Mag": mags,
                    }
                elif "/MagInfo/TotalMagOnAtomX" in hf.keys():  # noncollinear
                    magx = np.asarray(hf.get("/MagInfo/TotalMagOnAtomX"))  # Natom x 1
                    magy = np.asarray(hf.get("/MagInfo/TotalMagOnAtomY"))  # Natom x 1
                    magz = np.asarray(hf.get("/MagInfo/TotalMagOnAtomZ"))  # Natom x 1
                    # iNoncollinear = True
                    D_mag_fix = {
                        "Mag_x": np.repeat(magx[np.newaxis, :], Nstep, axis=0).tolist(),
                        "Mag_y": np.repeat(magy[np.newaxis, :], Nstep, axis=0).tolist(),
                        "Mag_z": np.repeat(magz[np.newaxis, :], Nstep, axis=0).tolist(),
                    }
                else:
                    mag = np.zeros(shape=(Natom, 1))
                    mags = np.repeat(mag[np.newaxis, :], Nstep, axis=0).tolist()
                    D_mag_fix = {
                        "Mag": mags,
                    }

            except Exception as e:
                if str(e):  # ignore empty AssertionError()
                    print(e)
                mag = np.zeros(shape=(Natom, 1))

            # repeat atomFixs of shape Natom x 3 to Nstep x Natom x 3
            Atomfixs = np.repeat(atomfix[np.newaxis, :], Nstep, axis=0).reshape(
                Nstep,
                Natom,
                3,
            )
            D_mag_fix.update({"Fix_x": Atomfixs[:, :, 0].tolist()})
            D_mag_fix.update({"Fix_y": Atomfixs[:, :, 1].tolist()})
            D_mag_fix.update({"Fix_z": Atomfixs[:, :, 2].tolist()})

            # repeat latticeFixs of shape 9 x 1 to Nstep x Natom x 9
            latticeFixs = (
                np.repeat(latticeFixs[np.newaxis, :], Nstep * Natom, axis=0)
                .reshape(Nstep, Natom, 9)
                .tolist()
            )
            D_mag_fix.update({"FixLattice": latticeFixs})

            if scaled:  # Fractional coordinates
                for k, ind in enumerate(indices):  # Step count
                    poses[k] = wrapped_poses[k]
            else:  # Cartesian coordinates
                for k, ind in enumerate(indices):  # Steps
                    poses[k] = wrapped_poses[k] @ lattices[k]

        elif "/RelaxedStructure" in hf.keys():  # Latest NEB chain
            raise NotImplementedError("neb.h5 is not supported yet")
        elif "/UnitAtomInfo" in hf.keys():  # phonon only reads unit cell information
            raise NotImplementedError("phonon.h5 is not supported yet")

        else:  # rho, potential, elf, pcharge
            hfDict = load_h5(absfile)
            s = _get_structure(hfDict, "/AtomInfo")
            elements = np.asarray(get_ele_from_h5(absfile), dtype=object)
            poses = [s.cart_coords]
            lattices = [s.lattice.matrix]
            Nstep = 1
            D_mag_fix = None

            logger.warning(
                "--> rho/potential/elf/pcharge.h5 has no mag or fix info,\n  you should manually set it if you are going to start new calculations..",
            )

    elif absfile.endswith(".json"):
        logger.warning(
            "float number in json has precision of 4 digits by default, which may cause inconsistency with h5/log file, you may use io.jsonPrec to adjust the precision",
            category=UserWarning,
        )
        with open(absfile) as f:
            from json import load

            data = load(f)  # Load JSON file
        # decide the task type by checking the internal keys
        if "AtomInfo" in data:  # single-step task
            s = _get_structure_json(data["AtomInfo"])
            elements = [str(i) for i in s.species]
            poses = [s.cart_coords]
            lattices = [s.lattice.matrix]
            Nstep = 1
            D_mag_fix = None

        elif "UnitAtomInfo" in data:  # phonon task
            raise NotImplementedError("Read from phonon.json is not supported yet.")
        elif "IniFin" in data:  # neb.json
            raise NotImplementedError("Read from neb.json is not supported yet.")
        elif "WannierInfo" in data:
            raise NotImplementedError("wannier.json has no structure info!")

        else:  # multi-steps task
            if "Structures" in data:
                Total_step = len(data["Structures"])  # aimd.json
            else:
                Total_step = len(data)  # relax.json, neb01.json

            if ele is not None and ai is not None:
                raise ValueError("Cannot specify both ele and ai")
            # Number of steps
            if si is not None:
                if isinstance(si, int):  # 1
                    indices = [si]

                elif isinstance(si, list) or isinstance(ai, np.ndarray):  # [1,2,3]
                    indices = si

                elif isinstance(si, str):  # ':', '-3:'
                    indices = __parse_indices(si, Total_step)

                else:
                    raise ValueError("si=%s is invalid" % si)

                Nstep = len(indices)
            else:
                Nstep = Total_step
                indices = list(range(1, Nstep + 1))  # [1,Nstep+1)

            # Pre-read the total list of all elements, which will not change with the number of steps and will not "combine like terms"
            if "Structures" in data:
                Nele = len(data["Structures"][0]["Atoms"])  # relax.json
                total_elements = np.empty(
                    shape=(Nele), dtype=object
                )  # Unmerged elements list
                for i in range(Nele):
                    element = data["Structures"][0]["Atoms"][i]["Element"]
                    total_elements[i] = element
            else:
                if "Atoms" not in data[0]:
                    raise NotImplementedError("nebXX.json has no structure info!")
                Nele = len(data[0]["Atoms"])
                total_elements = np.empty(
                    shape=(Nele),
                    dtype=object,
                )  # Unmerged elements list
                for i in range(Nele):
                    element = data[0]["Atoms"][i]["Element"]
                    total_elements[i] = element

            Natom = len(total_elements)

            # Starting to read the unit cell and atomic positions
            # Select the structure based on the element's index in data['Structures']['%d' % index]['Atoms']
            if ele is not None:  # User specifies certain elements
                location = []
                if isinstance(ele, str):  # Single element symbol, e.g., 'Fe'
                    ele_list = list(ele)
                # List of multiple element symbols, such as ['Fe', 'O']
                elif isinstance(ele, (list, np.ndarray)):
                    ele_list = ele
                else:
                    raise TypeError("ele=%s is invalid" % ele)
                for e in ele_list:
                    location.append(np.where(total_elements == e)[0])
                location = np.concatenate(location)

            elif (
                ai is not None
            ):  # If the user specifies an atomic number, the element list should also be filtered accordingly
                if isinstance(ai, int):  # 1
                    ais = [ai]
                elif isinstance(ai, (list, np.ndarray)):  # [1,2,3]
                    ais = ai
                elif isinstance(ai, str):  # ':', '-3:'
                    ais = __parse_indices(ai, Natom)
                else:
                    raise ValueError("ai=%s is invalid" % ai)
                ais = [
                    i - 1 for i in ais
                ]  # Python counts from 0, but users count from 1
                location = ais
                # read lattices and poses

            else:  # If none are specified
                location = list(range(Natom))

            # Elements list that meets user requirements
            elements = total_elements[location]

            # Nstep x Natom x 3, positions are all fractional
            poses = np.empty(shape=(len(indices), len(elements), 3))
            lattices = np.empty(shape=(Nstep, 3, 3))  # Nstep x 3 x 3
            mags = []  # Nstep x Natom x ?
            Atomfixs = []  # Nstep x Natom x 1
            LatFixs = []  # Nstep x Natom x 9

            if "Structures" in data:  # aimd
                for i, ind in enumerate(indices):  # for every ionic step
                    lat = data["Structures"][ind - 1]["Lattice"]
                    lattices[i] = np.asarray(lat).reshape(3, 3)
                    mag_for_each_step = []
                    fix_for_each_step = []
                    if "FixLattice" in data["Structures"][ind - 1]:
                        fixlat_raw = data["Structures"][ind - 1]["FixLattice"]
                    else:
                        fixlat_raw = []
                    if fixlat_raw == []:
                        fixlat_raw = np.full((9, 1), fill_value=False).tolist()
                    fixlat_str = [
                        "True" if _v is True else "False" for _v in fixlat_raw
                    ]
                    fixlat_arr = np.asarray(fixlat_str).reshape(9, 1)
                    # repeat fixlat for each atom
                    fixlat = np.repeat(fixlat_arr, Natom, axis=1).T.tolist()
                    LatFixs.append(fixlat)
                    for j, sli in enumerate(location):
                        ati = data["Structures"][ind - 1]["Atoms"][sli]
                        poses[i, j, :] = ati["Position"][:]

                        mag_for_each_atom = ati["Mag"][:]
                        if mag_for_each_atom == []:
                            mag_for_each_atom = [0.0]
                        mag_for_each_step.append(mag_for_each_atom)

                        fix_for_each_atom = ati["Fix"][:]
                        if fix_for_each_atom == []:
                            fix_for_each_atom = ["False"]
                        fix_for_each_step.append(fix_for_each_atom)

                    mags.append(mag_for_each_step)
                    Atomfixs.append(fix_for_each_step)
                    if not scaled:
                        poses[i] = np.dot(poses[i], lattices[i])

            else:  # relax, neb01
                logger.warning(
                    "mag and fix info are not available for relax.json and nebXX.json yet, trying read info...",
                    category=UserWarning,
                )

                for i, ind in enumerate(indices):  # for every ionic step
                    lat = data[ind - 1]["Lattice"]
                    lattices[i] = np.asarray(lat).reshape(3, 3)
                    mag_for_each_step = []
                    fix_for_each_step = []
                    if "FixLattice" in data[ind - 1]:
                        fixlat_raw = data[ind - 1]["FixLattice"]
                        if fixlat_raw is None:
                            fixlat_raw = np.full((9, 1), fill_value=False).tolist()
                        fixlat_str = [
                            "True" if _v is True else "False" for _v in fixlat_raw
                        ]
                        fixlat_arr = np.asarray(fixlat_str).reshape(9, 1)
                        # repeat fixlat for each atom
                        fixlat = np.repeat(fixlat_arr, Natom, axis=1).T.tolist()
                    else:
                        fixlat = np.full((Natom, 9), fill_value=False).tolist()

                    LatFixs.append(fixlat)
                    for j, sli in enumerate(location):
                        ati = data[ind - 1]["Atoms"][sli]
                        poses[i, j, :] = ati["Position"][:]

                        mag_for_each_atom = ati["Mag"][:]
                        if mag_for_each_atom == []:
                            mag_for_each_atom = [0.0]
                        mag_for_each_step.append(mag_for_each_atom)

                        fix_for_each_atom = ati["Fix"][:]
                        if fix_for_each_atom == []:
                            fix_for_each_atom = ["False"]
                        fix_for_each_step.append(fix_for_each_atom)

                    mags.append(mag_for_each_step)
                    Atomfixs.append(fix_for_each_step)
                    if not scaled:
                        poses[i] = np.dot(poses[i], lattices[i])

            elements = elements.tolist()
            Mags = np.asarray(mags).tolist()  # (Nstep, Natom, ?) or (Nstep, 0,)

            D_mag_fix = {"Mag": Mags, "Fix": Atomfixs, "LatticeFixs": LatFixs}

    else:
        raise ValueError(
            "get_sinfo function only accept datafile of .h5 / .json format!",
        )

    return Nstep, elements, poses, lattices, D_mag_fix


def load_h5(dir_h5: str) -> dict:
    """Traverse and read data from h5 file, save it in dictionary format

    Use this function with caution, as it will read a lot of unnecessary data and take a long time.

    Parameters
    ----------
    dir_h5:
        Path to the h5 file

    Returns
    -------
    data:
        Data dictionary

    Examples
    --------
    >>> from dspawpy.io.read import load_h5
    >>> data = load_h5(dir_h5='tests/2.2/scf.h5')
    >>> data.keys()
    dict_keys(['/AtomInfo/CoordinateType', '/AtomInfo/Elements', '/AtomInfo/Grid', '/AtomInfo/Lattice', '/AtomInfo/Position', '/Eigenvalue/CBM/BandIndex', '/Eigenvalue/CBM/Energy', '/Eigenvalue/CBM/Kpoint', '/Eigenvalue/NumberOfBand', '/Eigenvalue/Spin1/BandEnergies', '/Eigenvalue/Spin1/Kpoints/Coordinates', '/Eigenvalue/Spin1/Kpoints/Grid', '/Eigenvalue/Spin1/Kpoints/NumberOfKpoints', '/Eigenvalue/Spin1/Occupation', '/Eigenvalue/VBM/BandIndex', '/Eigenvalue/VBM/Energy', '/Eigenvalue/VBM/Kpoint', '/Electron', '/Energy/EFermi', '/Energy/TotalEnergy', '/Energy/TotalEnergy0', '/Force/ForceOnAtoms', '/Stress/Direction', '/Stress/Pressure', '/Stress/Stress', '/Stress/Total', '/Structures/FinalStep', '/Structures/Step-1/Lattice', '/Structures/Step-1/Position'])

    """

    def get_names(key, h5_object):
        names.append(h5_object.name)

    def is_dataset(name):
        for name_inTheList in names:
            if name_inTheList.find(name + "/") != -1:
                return False
        return True

    import numpy as np

    def get_data(key, h5_object):
        if is_dataset(h5_object.name):
            _data = np.asarray(h5_object)
            if _data.dtype == "|S1":  # Convert to string and split by ";"
                byte2str = [str(bi, "utf-8") for bi in _data]
                string = ""
                for char in byte2str:
                    string += char
                _data = np.asarray([elem for elem in string.strip().split(";")])
            # "/group1/group2/.../groupN/dataset" : value
            data[h5_object.name] = _data.tolist()

    import os

    from h5py import File

    with File(os.path.abspath(dir_h5), "r") as fin:
        names = []
        data = {}
        fin.visititems(get_names)
        fin.visititems(get_data)

        return data


def __parse_indices(index: str, maxIndex: int) -> list:
    """Parse the atomic or structural index string input by the user

    Input:
        - index: User input atom index/element string, e.g. '1:3,5,7:10'
        - maxIndex: Maximum index, e.g., 10
    Output:
        - indices: List of parsed atomic numbers, e.g., [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    """
    assert ":" in index, (
        "If you don't want to slice the index, please enter an integer or a list"
    )
    blcs = index.split(",")
    indices = []
    for blc in blcs:
        if ":" in blc:  # Slicing
            low = blc.split(":")[0]
            if not low:
                low = 1  # Start from 1
            else:
                low = int(low)
                assert low > 0, "Index start at 1!"
            high = blc.split(":")[1]
            if not high:
                high = maxIndex
            else:
                high = int(high)
                assert high <= maxIndex, "Index too large!"

            for i in range(low, high + 1):
                indices.append(i)
        else:  # Single number
            indices.append(int(blc))
    return indices


def _get_lammps_non_orthogonal_box(lat: Sequence):
    """Calculate the box boundary parameters for input to LAMMPS, used for generating dump structure files

    Parameters
    ----------
    lat : np.ndarray
        Common non-triangular 3x3 matrices

    Returns
    -------
    box_bounds:
        Used for inputting box boundaries into LAMMPS

    """
    # https://docs.lammps.org/Howto_triclinic.html
    A = lat[0]
    B = lat[1]
    C = lat[2]
    import numpy as np

    assert np.cross(A, B).dot(C) > 0, "Lat is not right handed"

    # Convert a general 3x3 matrix to a standard upper triangular matrix
    alpha = np.arccos(np.dot(B, C) / (np.linalg.norm(B) * np.linalg.norm(C)))
    beta = np.arccos(np.dot(A, C) / (np.linalg.norm(A) * np.linalg.norm(C)))
    gamma = np.arccos(np.dot(A, B) / (np.linalg.norm(A) * np.linalg.norm(B)))

    ax = np.linalg.norm(A)
    a = np.asarray([ax, 0, 0])

    bx = np.linalg.norm(B) * np.cos(gamma)
    by = np.linalg.norm(B) * np.sin(gamma)
    b = np.asarray([bx, by, 0])

    cx = np.linalg.norm(C) * np.cos(beta)
    cy = (np.linalg.norm(B) * np.linalg.norm(C) - bx * cx) / by
    cz = np.sqrt(abs(np.linalg.norm(C) ** 2 - cx**2 - cy**2))
    c = np.asarray([cx, cy, cz])

    # triangluar matrix in lammmps cell format
    # note that in OVITO, it will be down-triangular one
    # lammps_lattice = np.asarray([a,b,c]).T

    # write lammps box parameters
    # https://docs.lammps.org/Howto_triclinic.html#:~:text=The%20inverse%20relationship%20can%20be%20written%20as%20follows
    lx = np.linalg.norm(a)
    xy = np.linalg.norm(b) * np.cos(gamma)
    xz = np.linalg.norm(c) * np.cos(beta)
    ly = np.sqrt(np.linalg.norm(b) ** 2 - xy**2)
    yz = (np.linalg.norm(b) * np.linalg.norm(c) * np.cos(alpha) - xy * xz) / ly
    lz = np.sqrt(np.linalg.norm(c) ** 2 - xz**2 - yz**2)

    # "The parallelepiped has its “origin” at (xlo,ylo,zlo) and is defined by 3 edge vectors starting from the origin given by a = (xhi-xlo,0,0); b = (xy,yhi-ylo,0); c = (xz,yz,zhi-zlo)."
    # Let the origin be at (0,0,0), then xlo = ylo = zlo = 0
    xlo = ylo = zlo = 0
    # https://docs.lammps.org/Howto_triclinic.html#:~:text=the%20LAMMPS%20box%20sizes%20(lx%2Cly%2Clz)%20%3D%20(xhi%2Dxlo%2Cyhi%2Dylo%2Czhi%2Dzlo)
    xhi = lx + xlo
    yhi = ly + ylo
    zhi = lz + zlo
    # https://docs.lammps.org/Howto_triclinic.html#:~:text=This%20bounding%20box%20is%20convenient%20for%20many%20visualization%20programs%20and%20is%20calculated%20from%20the%209%20triclinic%20box%20parameters%20(xlo%2Cxhi%2Cylo%2Cyhi%2Czlo%2Czhi%2Cxy%2Cxz%2Cyz)%20as%20follows%3A
    xlo_bound = xlo + np.min([0, xy, xz, xy + xz])
    xhi_bound = xhi + np.max([0, xy, xz, xy + xz])
    ylo_bound = ylo + np.min([0, yz])
    yhi_bound = yhi + np.max([0, yz])
    zlo_bound = zlo
    zhi_bound = zhi
    box_bounds = np.asarray(
        [
            [xlo_bound, xhi_bound, xy],
            [ylo_bound, yhi_bound, xz],
            [zlo_bound, zhi_bound, yz],
        ],
    )

    return box_bounds


def _get_total_dos(dos: dict):
    # h5 -> Dos Obj
    import numpy as np

    energies = np.asarray(dos["/DosInfo/DosEnergy"])
    from pymatgen.electronic_structure.core import Spin

    if dos["/DosInfo/SpinType"][0] != "collinear":
        densities = {Spin.up: np.asarray(dos["/DosInfo/Spin1/Dos"])}
    else:
        densities = {
            Spin.up: np.asarray(dos["/DosInfo/Spin1/Dos"]),
            Spin.down: np.asarray(dos["/DosInfo/Spin2/Dos"]),
        }

    efermi = dos["/DosInfo/EFermi"][0]

    from pymatgen.electronic_structure.dos import Dos

    return Dos(efermi, energies, densities)


def _get_total_dos_json(dos: dict):
    # json -> Dos Obj
    import numpy as np

    energies = np.asarray(dos["DosInfo"]["DosEnergy"])
    from pymatgen.electronic_structure.core import Spin

    if dos["DosInfo"]["SpinType"] != "collinear":
        densities = {Spin.up: np.asarray(dos["DosInfo"]["Spin1"]["Dos"])}
    else:
        densities = {
            Spin.up: np.asarray(dos["DosInfo"]["Spin1"]["Dos"]),
            Spin.down: np.asarray(dos["DosInfo"]["Spin2"]["Dos"]),
        }
    efermi = dos["DosInfo"]["EFermi"]
    from pymatgen.electronic_structure.dos import Dos

    return Dos(efermi, energies, densities)


def _get_complete_dos(dos: dict):
    # h5 -> CompleteDos Obj
    total_dos = _get_total_dos(dos)
    structure = _get_structure(dos, "/AtomInfo")
    N = len(structure)
    pdos = [{} for i in range(N)]
    number_of_spin = 2 if dos["/DosInfo/SpinType"][0] == "collinear" else 1

    from pymatgen.electronic_structure.core import Orbital, Spin

    for i in range(number_of_spin):
        spin_key = "Spin" + str(i + 1)
        spin = Spin.up if i == 0 else Spin.down
        if dos["/DosInfo/Project"][0]:
            atomindexs = dos["/DosInfo/" + spin_key + "/ProjectDos/AtomIndexs"][0]
            orbitindexs = dos["/DosInfo/" + spin_key + "/ProjectDos/OrbitIndexs"][0]
            for atom_index in range(atomindexs):
                for orbit_index in range(orbitindexs):
                    orbit_name = Orbital(orbit_index)
                    Contribution = dos[
                        "/DosInfo/"
                        + spin_key
                        + "/ProjectDos"
                        + str(atom_index + 1)
                        + "/"
                        + str(orbit_index + 1)
                    ]
                    if orbit_name in pdos[atom_index].keys():
                        pdos[atom_index][orbit_name].update({spin: Contribution})
                    else:
                        pdos[atom_index][orbit_name] = {spin: Contribution}

            pdoss = {structure[i]: pd for i, pd in enumerate(pdos)}
        else:
            pdoss = {}

    from pymatgen.electronic_structure.dos import CompleteDos

    return CompleteDos(structure, total_dos, pdoss)  # type: ignore


def _get_complete_dos_json(dos: dict):
    # json -> CompleteDos Obj
    total_dos = _get_total_dos_json(dos)
    structure = _get_structure_json(dos["AtomInfo"])
    N = len(structure)
    pdos = [{} for i in range(N)]
    number_of_spin = 2 if dos["DosInfo"]["SpinType"] == "collinear" else 1

    from pymatgen.electronic_structure.core import Orbital, Spin

    for i in range(number_of_spin):
        spin_key = "Spin" + str(i + 1)
        spin = Spin.up if i == 0 else Spin.down
        if dos["DosInfo"]["Project"]:
            project = dos["DosInfo"][spin_key]["ProjectDos"]
            for p in project:
                atom_index = p["AtomIndex"] - 1
                o = p["OrbitIndex"] - 1
                orbit_name = Orbital(o)
                if orbit_name in pdos[atom_index].keys():
                    pdos[atom_index][orbit_name].update({spin: p["Contribution"]})
                else:
                    pdos[atom_index][orbit_name] = {spin: p["Contribution"]}
            pdoss = {structure[i]: pd for i, pd in enumerate(pdos)}
        else:
            pdoss = {}

    from pymatgen.electronic_structure.dos import CompleteDos

    return CompleteDos(structure, total_dos, pdoss)  # type: ignore


def _get_structure(hdf5: dict, key: str):
    """For single-step task"""
    # load_h5 -> Structure Obj
    import numpy as np

    lattice = np.asarray(hdf5[key + "/Lattice"]).reshape(3, 3)
    elements = hdf5[key + "/Elements"]
    positions = hdf5[key + "/Position"]
    coords = np.asarray(positions).reshape(-1, 3)
    is_direct = hdf5[key + "/CoordinateType"][0] == "Direct"
    import re

    elements = [re.sub(r"_", "", e) for e in elements]

    from pymatgen.core.structure import Structure

    return Structure(lattice, elements, coords, coords_are_cartesian=(not is_direct))


def _get_structure_json(atominfo: dict):
    """For single-step task"""
    import numpy as np

    lattice = np.asarray(atominfo["Lattice"]).reshape(3, 3)
    elements = []
    positions = []
    for atom in atominfo["Atoms"]:
        elements.append(atom["Element"])
        positions.extend(atom["Position"])

    coords = np.asarray(positions).reshape(-1, 3)
    is_direct = atominfo["CoordinateType"] == "Direct"
    import re

    elements = [re.sub(r"_", "", e) for e in elements]

    from pymatgen.core.structure import Structure

    return Structure(lattice, elements, coords, coords_are_cartesian=(not is_direct))


def _get_band_data_h5(band: dict, iwan: bool = False, zero_to_efermi: bool = False):
    if iwan:
        bd = "WannBandInfo"
    else:
        bd = "BandInfo"
    number_of_band = band[f"/{bd}/NumberOfBand"][0]
    number_of_kpoints = band[f"/{bd}/NumberOfKpoints"][0]
    if band[f"/{bd}/SpinType"][0] != "collinear":
        number_of_spin = 1
    else:
        number_of_spin = 2

    symmetry_kPoints_index = band[f"/{bd}/SymmetryKPointsIndex"]

    efermi = band[f"/{bd}/EFermi"][0]
    eigenvals = {}
    import numpy as np
    from pymatgen.electronic_structure.core import Spin

    for i in range(number_of_spin):
        spin_key = "Spin" + str(i + 1)
        spin = Spin.up if i == 0 else Spin.down

        if f"/{bd}/" + spin_key + "/BandEnergies" in band:
            data = band[f"/{bd}/" + spin_key + "/BandEnergies"]
        elif f"/{bd}/" + spin_key + "/Band" in band:
            data = band[f"/{bd}/" + spin_key + "/Band"]
        else:
            raise KeyError("Band key error")
        band_data = np.asarray(data).reshape((number_of_kpoints, number_of_band)).T

        if zero_to_efermi:
            eigenvals[spin] = band_data - efermi
        else:
            eigenvals[spin] = band_data

    kpoints = np.asarray(band[f"/{bd}/CoordinatesOfKPoints"]).reshape(
        number_of_kpoints,
        3,
    )

    structure = _get_structure(band, "/AtomInfo")
    labels_dict = {}

    for i, s in enumerate(band[f"/{bd}/SymmetryKPoints"]):
        labels_dict[s] = kpoints[symmetry_kPoints_index[i] - 1]

    # read projection data
    projections = None
    if f"/{bd}/IsProject" in band:
        if band[f"/{bd}/IsProject"][0]:
            projections = {}
            number_of_orbit = len(band[f"/{bd}/Orbit"])
            projection = np.zeros(
                (number_of_band, number_of_kpoints, number_of_orbit, len(structure)),
            )

            for i in range(number_of_spin):
                spin_key = "Spin" + str(i + 1)
                spin = Spin.up if i == 0 else Spin.down

                atomindexs = band[f"/{bd}/" + spin_key + "/ProjectBand/AtomIndex"][0]
                orbitindexs = band[f"/{bd}/" + spin_key + "/ProjectBand/OrbitIndexs"][0]
                for atom_index in range(atomindexs):
                    for orbit_index in range(orbitindexs):
                        project_data = band[
                            f"/{bd}/"
                            + spin_key
                            + "/ProjectBand/1/"
                            + str(atom_index + 1)
                            + "/"
                            + str(orbit_index + 1)
                        ]
                        projection[:, :, orbit_index, atom_index] = (
                            np.asarray(project_data)
                            .reshape((number_of_kpoints, number_of_band))
                            .T
                        )
                projections[spin] = projection

    if zero_to_efermi:
        efermi = 0  # set to 0

    return structure, kpoints, eigenvals, efermi, labels_dict, projections


def _get_band_data_json(
    band: dict,
    syst: Optional[dict] = None,
    iwan: bool = False,
    zero_to_efermi: bool = False,
):
    # syst is only required for wannier band structure
    if iwan:
        bd = "WannBandInfo"
        assert syst is not None, "syst is required for wannier band structure"
        efermi = syst["Energy"]["EFermi"]
        structure = _get_structure_json(syst["AtomInfo"])
    else:
        bd = "BandInfo"
        if "EFermi" in band[bd]:
            efermi = band[bd]["EFermi"]
        else:
            logger.warning("EFermi not found in band data, set to 0")
            efermi = 0
        structure = _get_structure_json(band["AtomInfo"])

    number_of_band = band[bd]["NumberOfBand"]
    number_of_kpoints = band[bd]["NumberOfKpoints"]
    # ! wannier.json has no SpinType key
    # if band[bd]["SpinType"][0] != "collinear":
    if "Spin2" not in band[bd]:
        number_of_spin = 1
    else:
        number_of_spin = 2

    symmetry_kPoints_index = band[bd]["SymmetryKPointsIndex"]
    eigenvals = {}
    import numpy as np
    from pymatgen.electronic_structure.core import Spin

    for i in range(number_of_spin):
        spin_key = "Spin" + str(i + 1)
        spin = Spin.up if i == 0 else Spin.down

        if "BandEnergies" in band[bd][spin_key]:
            data = band[bd][spin_key]["BandEnergies"]
        elif "Band" in band[bd][spin_key]:
            data = band[bd][spin_key]["Band"]
        else:
            raise KeyError("Band key error")

        band_data = np.asarray(data).reshape((number_of_kpoints, number_of_band)).T

        if zero_to_efermi:
            eigenvals[spin] = band_data - efermi

        else:
            eigenvals[spin] = band_data

    kpoints = np.asarray(band[bd]["CoordinatesOfKPoints"]).reshape(number_of_kpoints, 3)

    labels_dict = {}

    for i, s in enumerate(band[bd]["SymmetryKPoints"]):
        labels_dict[s] = kpoints[symmetry_kPoints_index[i] - 1]

    # read projection data
    projections = None
    if "IsProject" in band[bd].keys():
        if band[bd]["IsProject"]:
            projections = {}
            number_of_orbit = len(band[bd]["Orbit"])
            projection = np.zeros(
                (number_of_band, number_of_kpoints, number_of_orbit, len(structure)),
            )

            for i in range(number_of_spin):
                spin_key = "Spin" + str(i + 1)
                spin = Spin.up if i == 0 else Spin.down

                data = band[bd][spin_key]["ProjectBand"]
                for d in data:
                    orbit_index = d["OrbitIndex"] - 1
                    atom_index = d["AtomIndex"] - 1
                    project_data = d["Contribution"]
                    projection[:, :, orbit_index, atom_index] = (
                        np.asarray(project_data)
                        .reshape((number_of_kpoints, number_of_band))
                        .T
                    )
                projections[spin] = projection

    if zero_to_efermi:
        logger.warning("Setting efemi to 0 because zero_to_efermi is True")
        efermi = 0  # set to 0

    return structure, kpoints, eigenvals, efermi, labels_dict, projections


def _get_phonon_band_data_h5(band: dict):
    import numpy as np

    number_of_band = band["/BandInfo/NumberOfBand"][0]
    number_of_qpoints = band["/BandInfo/NumberOfQPoints"][0]
    symmmetry_qpoints = band["/BandInfo/SymmetryQPoints"]
    symmetry_qPoints_index = band["/BandInfo/SymmetryQPointsIndex"]
    qpoints = np.asarray(band["/BandInfo/CoordinatesOfQPoints"]).reshape(
        number_of_qpoints,
        3,
    )
    if "/SupercellAtomInfo/CoordinateType" in band:
        structure = _get_structure(band, "/SupercellAtomInfo")
    else:
        structure = _get_structure(band, "/AtomInfo")

    spin_key = "Spin1"
    if "/BandInfo/" + spin_key + "/BandEnergies" in band:
        data = band["/BandInfo/" + spin_key + "/BandEnergies"]
    elif "/BandInfo/" + spin_key + "/Band" in band:
        data = band["/BandInfo/" + spin_key + "/Band"]
    else:
        raise KeyError("Band key error")
    frequencies = np.asarray(data).reshape((number_of_qpoints, number_of_band)).T

    return symmmetry_qpoints, symmetry_qPoints_index, qpoints, structure, frequencies


def _get_phonon_band_data_json(band: dict):
    import numpy as np

    number_of_band = band["BandInfo"]["NumberOfBand"]
    number_of_qpoints = band["BandInfo"]["NumberOfQPoints"]

    symmmetry_qpoints = band["BandInfo"]["SymmetryQPoints"]
    symmetry_qPoints_index = band["BandInfo"]["SymmetryQPointsIndex"]
    qpoints = np.asarray(band["BandInfo"]["CoordinatesOfQPoints"]).reshape(
        number_of_qpoints,
        3,
    )
    if "SupercellAtomInfo" in band:
        structure = _get_structure_json(band["SupercellAtomInfo"])
    else:
        structure = _get_structure_json(band["AtomInfo"])

    spin_key = "Spin1"
    if "BandEnergies" in band["BandInfo"][spin_key]:
        data = band["BandInfo"][spin_key]["BandEnergies"]
    elif "Band" in band["BandInfo"][spin_key]:
        data = band["BandInfo"][spin_key]["Band"]
    else:
        raise KeyError("Band key error")
    frequencies = np.asarray(data).reshape((number_of_qpoints, number_of_band)).T

    return symmmetry_qpoints, symmetry_qPoints_index, qpoints, structure, frequencies



import threading
from typing import Literal

from loguru import logger
from prompt_toolkit import prompt

from dspawpy import auto_test_cli
from dspawpy.cli import list_ds, pc
from dspawpy.cli.menu_prompts import Dio, Dparameter, Dresponse


def s13_1(language: Literal["EN", "CN"]) -> dict:
    """Adsorbate"""
    if auto_test_cli:
        from dspawpy.io.utils import getTSads as cli_getTSads

        for D in list_ds:
            if D["menu"] == 131:
                cli_getTSads(D["inf"], float(D["T"] or 298.15))

        D = {}
    else:

        def imp():
            global getTSads
            from dspawpy.io.utils import getTSads

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 131
        D["inf"] = prompt(Dio[language]["txt"], completer=pc)
        D["T"] = float(input(Dparameter[language][6]) or 298.15)
        import_thread.join()
        logger.info(Dresponse[language][16])

        getTSads(D["inf"], D["T"])

    return D


def s13_2(language: Literal["EN", "CN"]) -> dict:
    """Ideal gas"""
    if auto_test_cli:
        from dspawpy.io.utils import getTSgas as cli_getTSgas

        for D in list_ds:
            if D["menu"] == 132:
                cli_getTSgas(
                    fretxt=D["inf"],
                    datafile=D["dataf"],
                    temperature=float(D["T"] or 298.15),
                    pressure=float(D["P"] or 101325.0),
                )

        D = {}
    else:

        def imp():
            global getTSgas
            from dspawpy.io.utils import getTSgas

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D = {}
        D["menu"] = 132
        D["inf"] = prompt(Dio[language]["txt"], completer=pc)
        D["dataf"] = prompt(Dio[language]["datafile"], completer=pc)
        D["T"] = float(input(Dparameter[language][6]) or 298.15)
        D["P"] = float(input(Dparameter[language][7]) or 101325.0)
        import_thread.join()
        logger.info(Dresponse[language][16])

        getTSgas(
            fretxt=D["inf"],
            datafile=D["dataf"],
            temperature=D["T"],
            pressure=D["P"],
        )
    return D



import json
import os

from prompt_toolkit.completion import (
    FuzzyCompleter,
    PathCompleter,
)

from dspawpy import auto_test_cli

pc = FuzzyCompleter(PathCompleter(expanduser=True))  # path completion
pc_h5 = FuzzyCompleter(
    PathCompleter(expanduser=True, file_filter=lambda p: p.endswith(".h5"))
)  # path completion

list_ds = []
if auto_test_cli:
    json_file = os.path.abspath("cli_input.json")
    if os.path.isfile(json_file):
        with open(json_file, "r") as fin:
            list_ds: list[dict] = [
                json.loads(line) for line in fin.readlines() if line.strip()
            ]
    else:
        raise FileNotFoundError(
            f"dspawpy_cli_test set to auto but {json_file} not found"
        )



import threading
from typing import Literal

from loguru import logger
from prompt_toolkit import prompt

from dspawpy import auto_test_cli
from dspawpy.cli import list_ds, pc
from dspawpy.cli.auxiliary import get_input, get_inputs, get_lims, save_figure_dpi300
from dspawpy.cli.menu_prompts import Dio, Dparameter, Dresponse, Dselect


def pre_ele_band(language: Literal["EN", "CN"]):
    def imp():
        global get_band_data, BSPlotter
        from pymatgen.electronic_structure.plotter import BSPlotter

        from dspawpy.io.read import get_band_data

    import_thread = threading.Thread(target=imp)
    import_thread.start()

    D = {}
    D["inf"] = prompt(Dio[language]["band"], completer=pc)
    D["ylims"] = get_lims(Dparameter[language][10], language)
    D["figure"] = prompt(Dio[language]["figure"], completer=pc) or "band.png"

    import_thread.join()
    logger.info(Dresponse[language][16])

    return D, BSPlotter, get_band_data


def pre_ele_pband(language: Literal["EN", "CN"]):
    def imp():
        global get_band_data, BSPlotterProjected
        from pymatgen.electronic_structure.plotter import BSPlotterProjected

        from dspawpy.io.read import get_band_data

    import_thread = threading.Thread(target=imp)
    import_thread.start()

    D = {}
    D["inf"] = prompt(Dio[language]["pband"], completer=pc)
    D["ylims"] = get_lims(Dparameter[language][10], language)
    D["figure"] = prompt(Dio[language]["figure"], completer=pc) or "band_projected.png"

    import_thread.join()
    logger.info(Dresponse[language][16])

    return D, BSPlotterProjected, get_band_data


def s4_1(language: Literal["EN", "CN"]) -> dict:
    """Normal band structure"""
    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import BSPlotter as cli_BSPlotter

        from dspawpy.io.read import get_band_data as cli_get_band_data

        for D in list_ds:
            band_data = cli_get_band_data(D["inf"], zero_to_efermi=True)
            is_metal = band_data.is_metal()

            if not is_metal and D["shift"] == "y":
                band_data = cli_get_band_data(D["inf"], zero_to_efermi=True)
                bsp = cli_BSPlotter(band_data)
                bsp.get_plot(False, ylim=D["ylims"])
            else:
                bsp = cli_BSPlotter(band_data)
                bsp.get_plot(True, ylim=D["ylims"])
            save_figure_dpi300(D["figure"])

        D = {}
    else:
        D, BSPlotter, get_band_data = pre_ele_band(language)
        D["menu"] = 41

        band_data = get_band_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        bsp = BSPlotter(band_data)
        logger.debug("plotter initialized")

        is_metal = band_data.is_metal()
        if is_metal:
            bsp.get_plot(ylim=D["ylims"])
        else:
            D["shift"] = get_input(Dselect[language][3], ["y", "n"])
            if D["shift"] == "y":
                from pymatgen.electronic_structure.plotter import BSPlotter

                from dspawpy.io.read import get_band_data

                band_data = get_band_data(D["inf"], zero_to_efermi=True)
                bsp = BSPlotter(band_data)
                bsp.get_plot(False, ylim=D["ylims"])
            else:
                bsp.get_plot(ylim=D["ylims"])

        save_figure_dpi300(D["figure"])
    return D


def s4_2(language: Literal["EN", "CN"]) -> dict:
    """Plot the band projection for each element separately, with the size of data points indicating the contribution of that element to the orbital"""
    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import (
            BSPlotterProjected as cli_BSPlotterProjected,
        )

        from dspawpy.io.read import get_band_data as cli_get_band_data

        for D in list_ds:
            band_data = cli_get_band_data(D["inf"], zero_to_efermi=False)
            is_metal = band_data.is_metal()
            if not is_metal and D["shift"] == "y":
                band_data = cli_get_band_data(D["inf"], zero_to_efermi=True)
                bsp = cli_BSPlotterProjected(band_data)
                bsp.get_elt_projected_plots(ylim=D["ylims"])
            else:
                bsp = cli_BSPlotterProjected(band_data)
                bsp.get_elt_projected_plots(ylim=D["ylims"])

            save_figure_dpi300(D["figure"])
        D = {}
    else:
        D, BSPlotterProjected, get_band_data = pre_ele_pband(language)
        D["menu"] = 42
        band_data = get_band_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        bsp = BSPlotterProjected(band_data)
        logger.debug("plotter initialized")

        is_metal = band_data.is_metal()
        if is_metal:
            bsp.get_elt_projected_plots(ylim=D["ylims"])
        else:
            D["shift"] = get_input(Dselect[language][3], ["y", "n"])
            if D["shift"] == "y":
                from pymatgen.electronic_structure.plotter import BSPlotterProjected

                from dspawpy.io.read import get_band_data

                band_data = get_band_data(D["inf"], zero_to_efermi=True)
                bsp = BSPlotterProjected(band_data)
                bsp.get_elt_projected_plots(ylim=D["ylims"])
            else:
                bsp.get_elt_projected_plots(ylim=D["ylims"])

        save_figure_dpi300(D["figure"])
    return D


def s4_3(language: Literal["EN", "CN"]) -> dict:
    """Band projection onto different orbitals of different elements"""
    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import (
            BSPlotterProjected as cli_BSPlotterProjected,
        )

        from dspawpy.io.read import get_band_data as cli_get_band_data

        for D in list_ds:
            band_data = cli_get_band_data(D["inf"], zero_to_efermi=False)
            is_metal = band_data.is_metal()
            if not is_metal and D["shift"] == "y":
                band_data = cli_get_band_data(D["inf"], zero_to_efermi=True)
                bsp = cli_BSPlotterProjected(band_data)
                bsp.get_projected_plots_dots(D["dictio"], False, ylim=D["ylims"])
            else:
                bsp = cli_BSPlotterProjected(band_data)
                bsp.get_projected_plots_dots(D["dictio"], False, ylim=D["ylims"])

            save_figure_dpi300(D["figure"])
        D = {}

    else:
        D, BSPlotterProjected, get_band_data = pre_ele_pband(language)
        D["menu"] = 43
        band_data = get_band_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        bsp = BSPlotterProjected(band_data)
        logger.debug("plotter initialized")

        banddatatructure = band_data.structure
        assert banddatatructure is not None
        logger.info(banddatatructure)
        es = banddatatructure.composition.elements
        D["dictio"] = {}

        from pymatgen.core import Element

        while True:
            _e = get_input(Dselect[language][4], [str(e) for e in es], allow_empty=True)
            if _e == "":
                break
            e = Element(_e)

            available_orbitals = ["s"]
            orbitals = e.atomic_orbitals
            assert isinstance(orbitals, dict)
            for o in orbitals:
                if "p" in o:
                    available_orbitals.append("p")
                elif "d" in o:
                    available_orbitals.append("d")
                elif "f" in o:
                    available_orbitals.append("f")
            unique_orbitals = list(set(available_orbitals))
            _o = get_input(Dselect[language][5], unique_orbitals)
            _os = _o.split(" ")
            dict_eo = {_e: _os}
            # update dictio
            D["dictio"].update(dict_eo)

        is_metal = band_data.is_metal()
        if is_metal:
            bsp.get_projected_plots_dots(D["dictio"])
        else:
            D["shift"] = get_input(Dselect[language][3], ["y", "n"])
            if D["shift"] == "y":
                from pymatgen.electronic_structure.plotter import BSPlotterProjected

                from dspawpy.io.read import get_band_data

                band_data = get_band_data(D["inf"], zero_to_efermi=True)
                bsp = BSPlotterProjected(band_data)
                bsp.get_projected_plots_dots(D["dictio"], False, ylim=D["ylims"])
            else:
                bsp.get_projected_plots_dots(D["dictio"], ylim=D["ylims"])

        save_figure_dpi300(D["figure"])
    return D


def s4_4(language: Literal["EN", "CN"]) -> dict:
    "Projects band structure onto different atomic orbitals for different atoms"
    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import (
            BSPlotterProjected as cli_BSPlotterProjected,
        )

        from dspawpy.io.read import get_band_data as cli_get_band_data

        for D in list_ds:
            band_data = cli_get_band_data(D["inf"], zero_to_efermi=False)
            is_metal = band_data.is_metal()
            if not is_metal and D["shift"] == "y":
                band_data = cli_get_band_data(D["inf"], zero_to_efermi=True)
                bsp = cli_BSPlotterProjected(band_data)
                bsp.get_projected_plots_dots_patom_pmorb(
                    D["dictio"],
                    D["dictpa"],
                    zero_to_efermi=False,
                    ylim=D["ylims"],
                )
            else:
                bsp = cli_BSPlotterProjected(band_data)
                bsp.get_projected_plots_dots_patom_pmorb(
                    D["dictio"],
                    D["dictpa"],
                    ylim=D["ylims"],
                )

            save_figure_dpi300(D["figure"])

        D = {}
    else:
        D, BSPlotterProjected, get_band_data = pre_ele_pband(language)
        D["menu"] = 44
        band_data = get_band_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        bsp = BSPlotterProjected(band_data)
        logger.debug("plotter initialized")

        banddatatructure = band_data.structure
        assert banddatatructure is not None
        logger.info(banddatatructure)
        sites = banddatatructure.sites
        ns = [str(i) for i in range(len(sites))]
        D["dictio"] = {}
        D["dictpa"] = {}
        while True:
            _n = get_input(Dselect[language][6], ns, allow_empty=True)
            if _n == "":
                break
            _e = sites[int(_n)].specie
            available_orbitals = ["s"]
            orbitals = _e.atomic_orbitals
            assert orbitals is not None
            for o in orbitals:
                if "p" in o:
                    available_orbitals.append("p")
                    available_orbitals.append("px")
                    available_orbitals.append("py")
                    available_orbitals.append("pz")
                elif "d" in o:
                    available_orbitals.append("d")
                    available_orbitals.append("dxy")
                    available_orbitals.append("dyz")
                    available_orbitals.append("dxz")
                    available_orbitals.append("dx2")
                    available_orbitals.append("dz2")
                elif "f" in o:
                    available_orbitals.append("f")
                    available_orbitals.append("f_3")
                    available_orbitals.append("f_2")
                    available_orbitals.append("f_1")
                    available_orbitals.append("f0")
                    available_orbitals.append("f1")
                    available_orbitals.append("f2")
                    available_orbitals.append("f3")

            unique_orbitals = list(set(available_orbitals))
            D["dictpa"].update({str(_e): [int(_n) + 2]})
            _os = get_inputs(Dselect[language][7], unique_orbitals)
            dict_eo = {str(_e): _os}
            # update dictio
            D["dictio"].update(dict_eo)

        logger.info(f"dictpa: {D['dictpa']}")
        logger.info(f"dictio: {D['dictio']}")

        is_metal = band_data.is_metal()
        if is_metal:
            bsp.get_projected_plots_dots_patom_pmorb(
                D["dictio"],
                D["dictpa"],
                ylim=D["ylims"],
            )
        else:
            D["shift"] = get_input(Dselect[language][3], ["y", "n"])
            if D["shift"] == "y":
                from pymatgen.electronic_structure.plotter import BSPlotterProjected

                from dspawpy.io.read import get_band_data

                band_data = get_band_data(D["inf"], zero_to_efermi=True)

                bsp = BSPlotterProjected(band_data)
                bsp.get_projected_plots_dots_patom_pmorb(
                    D["dictio"],
                    D["dictpa"],
                    zero_to_efermi=False,
                    ylim=D["ylims"],
                )
            else:
                bsp.get_projected_plots_dots_patom_pmorb(
                    D["dictio"],
                    D["dictpa"],
                    ylim=D["ylims"],
                )

        save_figure_dpi300(D["figure"])

    return D


def s4_5(language: Literal["EN", "CN"]) -> dict:
    """Handles unfolding processing."""

    if auto_test_cli:
        from dspawpy.plot import plot_bandunfolding as cli_plot_bandunfolding

        for D in list_ds:
            cli_plt = cli_plot_bandunfolding(D["inf"])
            cli_plt.ylim(D["ylims"])
            save_figure_dpi300(D["figure"], cli_plt)

        D = {}
    else:

        def imp():
            global plot_bandunfolding, plt
            import matplotlib.pyplot as plt

            from dspawpy.plot import plot_bandunfolding

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D, _, _ = pre_ele_pband(language)
        D["menu"] = 45
        import_thread.join()
        logger.info(Dresponse[language][16])

        plt = plot_bandunfolding(D["inf"])
        plt.ylim(D["ylims"])
        save_figure_dpi300(D["figure"], plt)

    return D


def s4_6(language: Literal["EN", "CN"]) -> dict:
    """Band-compare band structure comparison plot processing"""

    if auto_test_cli:
        from pymatgen.electronic_structure.plotter import BSPlotter as cli_BSPlotter

        from dspawpy.io.read import get_band_data as cli_get_band_data

        for D in list_ds:
            band_data = cli_get_band_data(D["inf"])
            bsp = cli_BSPlotter(band_data)
            if isinstance(D["infile2"], list):
                bd1 = cli_get_band_data(D["infile2"][0], D["infile2"][1])
            else:  # str
                bd1 = cli_get_band_data(D["infile2"])
            assert bd1 is not None, Dresponse[language][6]
            bsp = cli_BSPlotter(bs=bd1)
            bsp2 = cli_BSPlotter(bs=band_data)
            bsp.add_bs(bsp2._bs)
            bsp.get_plot(bs_labels=["wannier interpolated", "DFT"], ylim=D["ylims"])
            save_figure_dpi300(D["figure"])

        D = {}
    else:

        def imp():
            global get_band_data, BSPlotter
            from pymatgen.electronic_structure.plotter import BSPlotter

            from dspawpy.io.read import get_band_data

        import_thread = threading.Thread(target=imp)
        import_thread.start()
        D, _, get_band_data = pre_ele_pband(language)
        D["menu"] = 46
        band_data = get_band_data(D["inf"])
        logger.debug("got data, initializing plotter...")
        bsp = BSPlotterProjected(band_data)
        logger.debug("plotter initialized")

        D["infile2"] = prompt(Dio[language]["wband"], completer=pc)
        if D["infile2"].endswith(".json"):
            D["infile3"] = prompt(
                Dio[language]["sysjson"],
                completer=pc,
            )
            D["infile2"] = [D["infile2"], D["infile3"]]
        import_thread.join()
        logger.info(Dresponse[language][16])

        if isinstance(D["infile2"], list):
            bd1 = get_band_data(D["infile2"][0], D["infile2"][1])
        else:  # str
            bd1 = get_band_data(D["infile2"])
        assert bd1 is not None, Dresponse[language][6]
        bsp = BSPlotter(bs=bd1)
        bsp2 = BSPlotter(bs=band_data)
        bsp.add_bs(bsp2._bs)
        bsp.get_plot(bs_labels=["wannier interpolated", "DFT"], ylim=D["ylims"])
        save_figure_dpi300(D["figure"])

    return D
